"""
monobit.encoding.taggers - glyph tagging

(c) 2020--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import unicodedata
from pathlib import Path
from importlib.resources import files
from functools import partial, wraps, cached_property

from .unicode import unicode_name, is_showable
from ..core.labels import to_label, Tag, Char, Codepoint
from ..base import reverse_dict
from .base import NotFoundError, Encoder, register_reader
from . import tables


def _get_label(labels, labeltype):
    """Get first label of given type from list."""
    for label in labels:
        label = to_label(label)
        if isinstance(label, labeltype):
            return label
    return labeltype()

_get_char = partial(_get_label, labeltype=Char)
_get_codepoint = partial(_get_label, labeltype=Codepoint)
_get_tag = partial(_get_label, labeltype=Tag)


class UnicodeNameTagger(Encoder):
    """Tag with unicode name."""

    def __init__(self):
        super().__init__(name='name')

    def tag(self, *labels):
        """Get unicode glyph name."""
        char = _get_char(labels)
        if not char:
            return Tag()
        return Tag(unicode_name(char.value))

    def char(self, *labels):
        """Get char from unicode glyph name."""
        tag = _get_tag(labels)
        elements = tag.value.split(',')
        try:
            return Char(''.join(
                unicodedata.lookup(elem.strip())
                for elem in elements
            ))
        except KeyError:
            return Char()


class DescriptionTagger(Encoder):
    """Tag with unicode names and characters."""

    def __init__(self):
        super().__init__(name='desc')

    def tag(self, *labels):
        """Get unicode glyph name."""
        char = _get_char(labels)
        if not char:
            return Tag()
        char = char.value
        name = unicode_name(char)
        if is_showable(char):
            return Tag('[{}] {}'.format(char, name))
        return Tag(name)


class CharTagger(Encoder):
    """Tag with unicode characters."""

    def __init__(self):
        super().__init__(name='char')

    def tag(self, *labels):
        """Get showable char."""
        char = _get_char(labels).value
        if is_showable(char):
            return Tag(char)
        return Tag()


class FallbackTagger(Encoder):
    """Algorithmically generate tags."""

    def __init__(self, unmapped='glyph{count:04}'):
        """Set up mapping."""
        super().__init__(name='fallback')
        self._pattern = unmapped
        self._count = -1

    def tag(self, *labels):
        """Get value from tagmap."""
        self._count += 1
        return Tag(self._pattern.format(
            count=self._count,
            char=_get_char(labels),
            codepoint=_get_codepoint(labels),
            tag=_get_tag(labels),
        ))


class CodepointTagger(Encoder):
    """Tag with codepoint numbers."""

    def __init__(self, prefix=''):
        """Create codepoint tagger with prefix"""
        super().__init__(name='codepoint')
        self._prefix = prefix

    def tag(self, *labels):
        """Get codepoint string."""
        cp = _get_codepoint(labels)
        if not cp:
            return Tag()
        return Tag(f'{self._prefix}{cp}')


class Tagmap(Encoder):
    """Tag on the basis of a mapping table."""

    def __init__(self, mapping, name='', fallback=None):
        """Set up mapping."""
        super().__init__(name=name)
        self._chr2tag = mapping
        self._fallback = fallback or FallbackTagger()

    @cached_property
    def _tag2chr(self):
        return reverse_dict(self._chr2tag)

    def tag(self, *labels):
        """Get value from tagmap."""
        char = _get_char(labels)
        try:
            return Tag(self._chr2tag[char])
        except KeyError:
            return self._fallback.tag(*labels)

    def char(self, *labels):
        """Get char value from tagmap."""
        for label in labels:
            if isinstance(label, Tag):
                try:
                    return Char(self._tag2chr[label.value])
                except KeyError:
                    pass
        return Char()

    @property
    def mapping(self):
        return {**self._chr2tag}

    def __or__(self, other):
        """Return encoding overlaid with all characters defined in right-hand side."""
        return Tagmap(mapping=self._chr2tag | other._chr2tag, name=f'{self.name}')

    def table(self):
        """Mapping table"""
        return '\n'.join(
            f'u+{ord(_k):04X}: "{_v}"' for _k, _v in self._chr2tag.items()
        )

class AdobeFallbackTagger(Encoder):
    """Fallback tagger following AGL conventions."""

    def __init__(self):
        super().__init__(name='adobe-fallback')

    def tag(self, *labels):
        """Construct a default tag for unmapped glyphs."""
        char = _get_char(labels)
        # following agl recommendation for naming sequences
        cps = (ord(_c) for _c in char)
        return Tag('_'.join(
            f'uni{_cp:04X}' if _cp < 0x10000 else f'u{_cp:06X}'
            for _cp in cps
        ))


class SGMLFallbackTagger(Encoder):
    """Fallback tagger following SGML conventions."""

    def __init__(self):
        super().__init__(name='sgml-fallback')

    def tag(self, *labels):
        """Construct a default tag for unmapped glyphs."""
        char = _get_char(labels)
        cps = (ord(_c) for _c in char)
        # joining numeric references by semicolons
        # note that each entity should really start with & and end with ; e.g. &eacute;
        return Tag(';'.join(f'#{_cp:X}' for _cp in cps))


###################################################################################################
# tag map format readers

def _tagmap_loader(fn):
    """Decorator for the shared parts of tagmap loaders."""

    @wraps(fn)
    def _load(name, path, *args, fallback, **kwargs):
        try:
            data = path.read_bytes()
        except EnvironmentError as exc:
            raise NotFoundError(f'Could not load tagmap file `{str(path)}`: {exc}')
        if not data:
            raise NotFoundError(f'No data in tagmap file `{str(path)}`')
        mapping = fn(data, *args, **kwargs)
        return Tagmap(mapping, name=name, fallback=fallback)

    return _load


@register_reader('tagmap')
@_tagmap_loader
def _read_tagmap(data, separator=':', comment='#', joiner=',', tag_column=0, unicode_column=1):
    """Read a tag map from file data."""
    chr2tag = {}
    for line in data.decode('utf-8').splitlines():
        if line.startswith(comment):
            continue
        columns = line.split(separator)
        if len(columns) > max(tag_column, unicode_column):
            tag = columns[tag_column]
            unicode_str = columns[unicode_column]
            try:
                char = ''.join(chr(int(_str, 16)) for _str in unicode_str.split(joiner))
            except ValueError:
                pass
            else:
                chr2tag[char] = tag
    return chr2tag



"""
monobit.storage.fontformats.raw.optiks - OPTIKS PCR fonts

(c) 2023--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging

from monobit.storage import loaders, savers, Magic
from monobit.core import Glyph, Font, Char
from monobit.base.struct import little_endian as le, bitfield
from monobit.base import UnsupportedError

from .raw import load_bitmap, save_bitmap
from monobit.storage.utils.limitations import ensure_single, ensure_charcell, make_contiguous


###############################################################################
# OPTIKS PCR - near-raw format
# http://fileformats.archiveteam.org/wiki/PCR_font
# http://cd.textfiles.com/simtel/simtel20/MSDOS/GRAPHICS/OKF220.ZIP
# OKF220.ZIP → OKFONTS.ZIP → FONTS.DOC - Has an overview of the format.
# > I have added 11 bytes to the head of the file
# > so that OPTIKS can identify it as a font file. The header has
# > a recognition pattern, OPTIKS version number and the size of
# > the font file.


_PCR_MAGIC_1 = b'KPG\1\1\x20\1'
_PCR_MAGIC_2 = b'KPG\1\2\x20\1'
_PCR_HEADER = le.Struct(
    magic='7s',
    # maybe it's a be uint16 of the file size, followed by the same size as le
    # anyway same difference
    height='uint8',
    zero='uint8',
    bytesize='uint16',
)

@loaders.register(
    name='pcr',
    magic=(_PCR_MAGIC_2, _PCR_MAGIC_1),
    patterns=('*.pcr',),
)
def load_pcr(instream):
    """Load an OPTIKS .PCR font."""
    header = _PCR_HEADER.read_from(instream)
    font = load_bitmap(instream, width=8, height=header.height, count=256)
    font = font.modify(source_format='Optiks PCR')
    return font


@savers.register(linked=load_pcr)
def save_pcr(fonts, outstream):
    """Save an OPTIKS .PCR font."""
    font = ensure_single(fonts)
    font = ensure_charcell(font)
    if font.cell_size.x != 8:
        raise UnsupportedError(
            'This format only supports 8xN character-cell fonts.'
        )
    font = make_contiguous(font, full_range=range(256), missing='space')
    header = _PCR_HEADER(
        magic=_PCR_MAGIC_2,
        height=font.cell_size.y,
        bytesize=font.cell_size.y * 0x100,
    )
    outstream.write(bytes(header))
    save_bitmap(outstream, font)



"""
monobit.encoding.definitions - character map definitions

(c) 2020--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import json
import logging
from pathlib import Path
from importlib.resources import files

from .registry import EncodingRegistry
from .charmaps import Unicode, EncoderLoader
from .taggers import (
    Tagmap, CharTagger, CodepointTagger,
    UnicodeNameTagger, DescriptionTagger,
    FallbackTagger, AdobeFallbackTagger, SGMLFallbackTagger,
)
from . import tables


def register_charmaps(charmaps):
    """Register charmap files"""
    for _name, _dict in json.loads((files(tables) / 'charmaps.json').read_text()).items():
        charmap = EncoderLoader(
            name=_name, filename=_dict['filename'],
            format=_dict.get('format', None),
            **_dict.get('kwargs', {}),
        )
        if 'range' in _dict:
            charmap = charmap.subset(_dict['range'])
        # overlays must be defined first
        for _overlay in _dict.get('overlays', ()):
            charmap |= charmaps.getter(_overlay)
        aliases = (_name, *_dict.get('aliases', ()))
        charmaps[aliases] = charmap


encodings = EncodingRegistry()

# unicode aliases
encodings['unicode', 'ucs', 'iso10646', 'iso10646-1'] = Unicode()

register_charmaps(encodings)


###############################################################################

for tagmap in (
        CharTagger(),
        CodepointTagger(),
        UnicodeNameTagger(),
        DescriptionTagger(),
    ):
    encodings[tagmap.name] = tagmap


encodings['adobe'] = EncoderLoader(
    'agl/aglfn.txt', name='adobe', format='tagmap',
    separator=';', unicode_column=0, tag_column=1,
    fallback=AdobeFallbackTagger()
)

encodings['sgml'] = EncoderLoader(
    'misc/SGML.TXT', name='sgml', format='tagmap',
    separator='\t', unicode_column=2,
    fallback=SGMLFallbackTagger()
)

# truetype mapping is adobe mapping *but* with .null for NUL
# https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6post.html
encodings['truetype'] = EncoderLoader(
    'agl/aglfn.txt', name='truetype', format='tagmap',
    separator=';', unicode_column=0, tag_column=1,
    fallback=AdobeFallbackTagger()
) | Tagmap({'\0': '.null'})



"""
monobit.base.binary - binary utilities

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""


def ceildiv(num, den):
    """Integer division, rounding up."""
    return -(-num // den)


def align(num, exp):
    """Round up to multiple of 2**exp."""
    mask = 2**exp - 1
    return (num + mask) & ~mask


def bytes_to_bits(byteseq, width=None):
    """
    Convert bytes/bytearray/sequence of int to tuple of bits.
    Note that bytes_to_pixels is the preferred alternative.
    """
    bitstr = bin(int.from_bytes(byteseq, 'big'))[2:].zfill(8 * len(byteseq))
    bits = tuple(_c == '1' for _c in bitstr)
    if width is None:
        return bits
    return bits[:width]


def int_to_bytes(in_int, byteorder='big'):
    """Convert integer to bytes."""
    return in_int.to_bytes(max(1, ceildiv(in_int.bit_length(), 8)), byteorder)


def bytes_to_int(in_bytes, byteorder='big'):
    """Convert integer to bytes."""
    return int.from_bytes(bytes(in_bytes), byteorder)


def reverse_by_group(bitseq, fill='0', group_size=8):
    """
    Reverse bits in every byte in string representation of binary
    Bit sequence is extended to end on byte boundary.
    """
    bitseq = bitseq.ljust(ceildiv(len(bitseq), group_size) * group_size, fill)
    args = [iter(bitseq)] * group_size
    bitseq = ''.join(''.join(_chunk[::-1]) for _chunk in zip(*args))
    return bitseq


###############################################################################
# bytes to pixels

# default string inklevels
INKLEVELS = {
    256: ''.join(chr(_i) for _i in range(256)),
    16: '0123456789abcdef',
    4: '0123',
    2: '01',
}


def bytes_to_pixels(byteseq, levels):
    """Convert bytes to pixels in level-specific representation."""
    if levels not in (2, 4, 16, 256):
        raise ValueError(f'Unsupported `levels` value: {levels}')
    if not byteseq:
        return ''
    if levels == 256:
        return byteseq.decode('latin-1')
    else:
        to_base = _base_converter(levels)
        bpp = (levels - 1).bit_length()
        pixels_per_byte = 8 // bpp
        return (
            to_base(int.from_bytes(byteseq, 'big'))
                .zfill(pixels_per_byte * len(byteseq))
        )


# base-4 conversion
_HEX_TO_QUAD = str.maketrans({
    '0': '00',
    '1': '01',
    '2': '02',
    '3': '03',
    '4': '10',
    '5': '11',
    '6': '12',
    '7': '13',
    '8': '20',
    '9': '21',
    'a': '22',
    'b': '23',
    'c': '30',
    'd': '31',
    'e': '32',
    'f': '33'
})


def _base_converter(base):
    """Converter to given base."""
    if base == 2:
        return (lambda _v: bin(_v)[2:])
    elif base == 4:
        # keep leading zero, we'll need it anyway
        return (lambda _v: hex(_v)[2:].translate(_HEX_TO_QUAD))
    elif base == 16:
        return (lambda _v: hex(_v)[2:])
    else:
        # we don't need the other bases
        raise ValueError(f'Unsupported base: {base}')



"""
monobit.storage.containerformats.apple - AppleSingle and AppleDouble containers

(c) 2021--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from pathlib import Path

from monobit.base.struct import big_endian as be
from monobit.base import FileFormatError
from ..streams import Stream
from ..magic import Magic
from ..base import containers
from ..containers import FlatFilterContainer


##############################################################################
# AppleSingle/AppleDouble container
# v1: see https://web.archive.org/web/20160304101440/http://kaiser-edv.de/documents/Applesingle_AppleDouble_v1.html
# v2: https://web.archive.org/web/20160303215152/http://kaiser-edv.de/documents/AppleSingle_AppleDouble.pdf
# the difference between v1 and v2 affects the file info sections
# not the resource fork which is what we care about

_APPLESINGLE_MAGIC = 0x00051600
_APPLEDOUBLE_MAGIC = 0x00051607


_APPLE_HEADER = be.Struct(
    magic='uint32',
    version='uint32',
    home_fs='16s',
    number_entities='uint16',
)
_APPLE_ENTRY = be.Struct(
    entry_id='uint32',
    offset='uint32',
    length='uint32',
)

# Entry IDs
_ID_DATA = 1
_ID_RESOURCE = 2
_ID_NAME = 3

_APPLE_ENTRY_TYPES = {
    1: 'data fork',
    2: 'resource fork',
    3: 'real name',
    4: 'comment',
    5: 'icon, b&w',
    6: 'icon, color',
    7: 'file info', # v1 only
    8: 'file dates info', # v2
    9: 'finder info',
    # the following are all v2
    10: 'macintosh file info',
    11: 'prodos file info',
    12: 'ms-dos file info',
    13: 'short name',
    14: 'afp file info',
    15: 'directory id',
}


class AppleContainer(FlatFilterContainer):


    def decode(self, name):
        """
        Decode data and resource fork from AppleSingle/AppeDouble container.
        """
        return super().decode(name)

    def encode(self, name):
        """
        Writing to AppleSingle/AppleDouble is not supported.
        """
        raise ValueError(
            'Writing to AppleSingle/AppleDouble is not supported.'
        )

    def decode_all(self, stream):
        """Parse an AppleSingle or AppleDouble file."""
        data = stream.read()
        header = _APPLE_HEADER.from_bytes(data)
        if header.magic == _APPLESINGLE_MAGIC:
            container = 'AppleSingle'
        elif header.magic == _APPLEDOUBLE_MAGIC:
            container = 'AppleDouble'
        else:
            raise FileFormatError('Not an AppleSingle or AppleDouble file.')
        entry_array = _APPLE_ENTRY.array(header.number_entities)
        entries = entry_array.from_bytes(data, _APPLE_HEADER.size)
        name, data_fork, rsrc_fork = '', b'', b''
        for i, entry in enumerate(entries):
            entry_type = _APPLE_ENTRY_TYPES.get(entry.entry_id, 'unknown')
            logging.debug(
                '%s container: entry #%d, %s [%d]',
                container, i, entry_type, entry.entry_id
            )
            if entry.entry_id == _ID_RESOURCE:
                rsrc_fork = data[entry.offset:entry.offset+entry.length]
            if entry.entry_id == _ID_DATA:
                data_fork = data[entry.offset:entry.offset+entry.length]
            if entry.entry_id == _ID_NAME:
                name = data[entry.offset:entry.offset+entry.length]
                name = name.decode('mac-roman')
        if name:
            return {
                f'data/{name}': data_fork,
                f'rsrc/{name}': rsrc_fork,
            }
        else:
            return {
                f'data': data_fork,
                f'rsrc': rsrc_fork,
            }


containers.register(
    name='apple1',
    magic=(
        _APPLESINGLE_MAGIC.to_bytes(4, 'big'),
    ),
    patterns=('*.as',),
)(AppleContainer)

containers.register(
    name='apple2',
    magic=(
        _APPLEDOUBLE_MAGIC.to_bytes(4, 'big'),
    ),
    # .adf, .rsrc - per http://fileformats.archiveteam.org/wiki/AppleDouble
    # ._<name> is OS X representation
    patterns=('*.adf', '*.rsrc', '._*'),
)(AppleContainer)



"""
monobit.storage.wrapperformats.bintext - text-encoded binary formats

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
import base64
import binascii
from io import BytesIO
from pathlib import Path

from ..streams import Stream, DelayedWriterStream
from ..base import encoders, decoders


@decoders.register(name='base64')
def decode_base64(instream):
    """Read from a base64-encoded binary file."""
    encoded_bytes = instream.read()
    data = base64.b64decode(encoded_bytes)
    name = Path(instream.name).stem
    return Stream.from_data(data, mode='r', name=name)


@encoders.register(linked=decode_base64)
def encode_base64(outstream, *, line_length:int=76):
    """
    Write to a base64-encoded binary file.

    line_length: length of each line of base64 encoded data (default: 76)
    """
    encode_func = _do_encode_base64
    name = Path(outstream.name).stem
    return DelayedWriterStream(
        outstream, encode_func, name,
        line_length=line_length,
    )


def _do_encode_base64(data, outstream, *, line_length):
    """Base64-encoded binary file."""
    encoded_bytes = base64.b64encode(data)
    bytesio = BytesIO(encoded_bytes)
    while True:
        line = bytesio.read(line_length)
        if not line:
            break
        outstream.write(line + b'\n')



@decoders.register(name='quopri')
def decode_quopri(instream):
    """Read from quoted-printable-encoded binary file."""
    encoded = instream.read()
    data = binascii.a2b_qp(encoded)
    name = Path(instream.name).stem
    return Stream.from_data(data, mode='r', name=name)


@encoders.register(linked=decode_quopri)
def encode_quopri(outstream, *, quote_tabs:bool=False):
    """
    Write to quoted-printable-encoded binary file.

    quote_tabs: encode tabs and spaces as QP (default: False)
    """
    encode_func = _do_encode_quopri
    name = Path(outstream.name).stem
    return DelayedWriterStream(
        outstream, encode_func, name,
        quote_tabs=quote_tabs,
    )


def _do_encode_quopri(data, outstream, *, quote_tabs):
    encoded = binascii.b2a_qp(data, quotetabs=quote_tabs, istext=False)
    outstream.write(encoded)



"""
monobit.storage.fontformats.sfont - SDL SFont format

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from itertools import groupby

from monobit.base import safe_import
Image = safe_import('PIL.Image')

from monobit.storage import loaders, savers
from monobit.base import FileFormatError, UnsupportedError
from monobit.core import Font, Glyph
from monobit.render import GlyphMap

from monobit.storage.utils.limitations import ensure_single
from .image import identify_inklevels

# see https://github.com/karlb/sfont


# _INDICATOR_RGBA = (255, 0, 255, 255)
_INDICATOR_RGB = (255, 0, 255)
_SFONT_RANGE = range(33, 127)


if Image:

    @loaders.register(
        name='sfont',
    )
    def load_sfont(instream):
        """
        Load font from SFont file.
        """
        image = Image.open(instream).convert('RGB')
        glyphs = []
        indicator = tuple(image.crop((0, 0, image.width, 1)).getdata())
        if len(set(indicator)) != 2:
            raise FileFormatError('Not an SFont image: missing indicator bar.')
        # we converted to RGB so we don't need to consider RGBA
        for rgb in set(indicator):
            if rgb != _INDICATOR_RGB:
                background = rgb
                break
        spritesheet = image.crop((0, 1, image.width, image.height))
        colours = spritesheet.getdata()
        # identify colour mode
        inklevels = identify_inklevels(colours, background)
        # find indicator lengths
        groups = tuple(
            (_clr, len(tuple(_g)))
            for _clr, _g in groupby(indicator)
        )
        # extract glyphs
        x = 0
        glyphs = []
        left = 0
        for i, (clr, length) in enumerate(groups):
            if clr == _INDICATOR_RGB:
                if i == 0:
                    left = length
                else:
                    if i == len(groups):
                        right = length
                    else:
                        right = length // 2
                    crop = spritesheet.crop(
                        (x, 0, x + left+width+right, spritesheet.height)
                    )
                    glyphs.append(
                        Glyph.from_vector(
                            tuple(crop.getdata()),
                            stride=crop.width,
                            inklevels=inklevels,
                            codepoint=min(_SFONT_RANGE) + i//2,
                            left_bearing=-left,
                            right_bearing=-right,
                        )
                    )
                    x += left+width+right + length%2
                    left = right
            else:
                width = length
        return Font(glyphs)


    @savers.register(linked=load_sfont)
    def save_sfont(fonts, outstream, *, image_format:str='png'):
        """
        Save font to SFont file.

        image_format: format of the image file. Default: `png`
        """
        font = ensure_single(fonts)
        font = font.equalise_horizontal()
        font = font.resample(codepoints=_SFONT_RANGE)
        glyphmap = GlyphMap()
        glyphmap.append_glyph(Glyph(), 0, 0)
        indicator = []
        right = 0
        x = 0
        for cp in _SFONT_RANGE:
            glyph = font.get_glyph(codepoint=cp)
            left = -glyph.left_bearing
            indicator_length = left
            if cp > min(_SFONT_RANGE):
                indicator_length += right + (right+left)%2 + 1
            x += indicator_length
            glyphmap.append_glyph(glyph, x, 0)
            width = glyph.advance_width
            x += width
            indicator.extend((_INDICATOR_RGB,) * indicator_length)
            indicator.extend(((0, 0, 0),) * width)
            right = -glyph.right_bearing
        if right:
            indicator.append(_INDICATOR_RGB * right)
        glyphmap.append_glyph(Glyph(), x, 0)
        glyph_image = glyphmap.as_image(
            ink=(255, 255, 255), paper=(0, 0, 0), border=(0, 0, 0)
        )
        image = Image.new(
            glyph_image.mode, (glyph_image.width, glyph_image.height+1)
        )
        image.paste(glyph_image, (0, 1))
        image.putdata(indicator)
        image.save(outstream, format=image_format)



"""
monobit.storage.fontformats.beos - BeOS Bitmap Font

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from itertools import accumulate

from monobit.base.binary import ceildiv
from monobit.base.struct import big_endian as be
from monobit.storage import loaders, savers
from monobit.core import Font, Glyph

from monobit.storage.utils.limitations import ensure_single


# http://www.eonet.ne.jp/~hirotsu/bin/bmf_format.txt

_HEADER = be.Struct(
    mark='4s',
    # > total size
    size='uint32',
    # > font-family-name size (not including the trailing null)
    ffnSize='uint16',
    # > font-style-name size (not including the trailing null)
    fsnSize='uint16',
    padding_0='10s',
    # > The number of characters that can be stored in the location-table -1
    ltMax='uint16',
    # > font-point (Bitmap fonts are enabled at this point number)
    point='uint16',
    # > 0x0300 (unknown)
    unknown_768='uint16',
    # unknown and differs per font; ProFont has it all null so that is valid?
    unknown='8s',
)

_LOCATION_ENTRY = be.Struct(
    pointer='uint32',
    code='uint16',
    reserved='uint16',
)

_GLYPH_DATA = be.Struct(
    unknown_0x4996b438 = 'uint32',
    unknown_0x4996b440 = 'uint32',
    left='int16',
    top='int16',
    right='int16',
    bottom='int16',
    width='float',
    maybe_height='float',
)

_BEOS_MAGIC = b'|Be;'


@loaders.register(
    name='beos',
    magic=(_BEOS_MAGIC,)
)
def load_beos(instream):
    """Load font from Be Bitmap Font file."""
    header = _HEADER.read_from(instream)
    familyName = instream.read(header.ffnSize+1)[:-1].decode('latin-1')
    styleName = instream.read(header.fsnSize+1)[:-1].decode('latin-1')
    logging.debug('header: %s', header)
    logging.debug('family: %s', familyName)
    logging.debug('style: %s', styleName)
    # hash table of pointers to glyphs, hashed by unicode codepoint
    location_table = (_LOCATION_ENTRY * (header.ltMax+1)).read_from(instream)
    location_dict = {_e.pointer: _e.code for _e in location_table}
    glyphs = []
    while instream.tell() < header.size:
        pointer = instream.tell()
        code = location_dict.get(pointer, None)
        glyph_data = _GLYPH_DATA.read_from(instream)
        # bitmap dimensions
        width = glyph_data.right - glyph_data.left + 1
        height = glyph_data.bottom - glyph_data.top + 1
        # 4 bits per pixel
        bytewidth = ceildiv(width * 4, 8)
        glyph_bytes = instream.read(height*bytewidth)
        glyphs.append(
            Glyph.from_bytes(
                glyph_bytes, width=width, height=height, bits_per_pixel=4,
                char=chr(code) if code is not None else code,
                right_bearing=round(glyph_data.width)-width-glyph_data.left,
                left_bearing=glyph_data.left,
                shift_up=-1-glyph_data.bottom,
                scalable_width=glyph_data.width,
            )
        )
    return Font(
        glyphs,
        encoding='unicode',
        family=familyName,
        subfamily=styleName,
        point_size=header.point,
    )


@savers.register(linked=load_beos)
def save_beos(fonts, outstream):
    """Save font to BeOS file."""
    font = ensure_single(fonts)
    font = font.label()
    # drop multi-codepoint sequences and unlabelled glyphs
    glyphs = tuple(_g for _g in font.glyphs if len(_g.char) == 1)
    # create header
    style_name = font.name[len(font.family):].strip()
    header = _HEADER(
        mark=_BEOS_MAGIC,
        # size='uint32',
        ffnSize=len(font.family),
        fsnSize=len(style_name),
        # ltMax is an assumption reproducing the value in my sample font
        ltMax = 2*(len(glyphs)-1)-1,
        point=font.point_size,
        unknown_768=0x300,
    )
    # create glyph table
    glyph_data = tuple(
        bytes(_GLYPH_DATA(
            unknown_0x4996b438 = 0x4996b438,
            unknown_0x4996b440 = 0x4996b440,
            left=_g.left_bearing,
            top=(-1-_g.shift_up) -_g.height + 1,
            right=_g.width + _g.left_bearing - 1,
            bottom=-1-_g.shift_up,
            width=_g.scalable_width,
            # maybe_height='float',
        ))
        for _g in glyphs
    )
    strike_offset = (
        _HEADER.size + header.ffnSize + 1 + header.fsnSize + 1
        + _LOCATION_ENTRY.size * (header.ltMax+1)
    )
    glyph_bytes = tuple(_g.as_bytes(bits_per_pixel=4) for _g in glyphs)
    offsets = accumulate(
        (len(_g) + len(_s) for _g, _s in zip(glyph_data, glyph_bytes)),
        initial=strike_offset,
    )
    # create location entries
    loc_entries = tuple(
        _LOCATION_ENTRY(pointer=_offs, code=ord(_g.char))
        for _g, _offs in zip(glyphs, offsets)
    )
    strike = b''.join(
        b''.join((_data, _bytes))
        for _data, _bytes in zip(glyph_data, glyph_bytes)
    )
    header.size = strike_offset + len(strike)
    # create hash table
    hashes = (
        ((ord(_g.char)>>2) ^ (ord(_g.char)<<3)) & header.ltMax
        for _g in glyphs
    )
    location_table = [None] * (header.ltMax+1)
    for entry, hash in zip(loc_entries, hashes):
        while location_table[hash] is not None:
            hash += 1
            if hash > header.ltMax:
                hash = 0
        location_table[hash] = entry
    location_table = (_LOCATION_ENTRY * (header.ltMax+1))(*(
        _entry if _entry else _LOCATION_ENTRY(pointer=0xffffffff)
        for _entry in location_table
    ))
    outstream.write(bytes(header))
    outstream.write(font.family.encode('latin-1', 'replace')+ b'\0')
    outstream.write(style_name.encode('latin-1', 'replace')+ b'\0')
    outstream.write(bytes(location_table))
    outstream.write(bytes(strike))



"""
monobit.storage.containerformats.uselibarchive - archive formats supported by libarchive

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from pathlib import Path


from monobit.base import safe_import
libarchive = safe_import('libarchive')
if libarchive:
    from libarchive.entry import FileType


from monobit.base import FileFormatError
from ..base import containers
from ..containers import FlatFilterContainer


if libarchive:

    class LibArchiveContainer(FlatFilterContainer):
        """Container for formats supported by libarchive."""

        libarchive_format = ''

        def decode(self, name):
            """Extract file from archive."""
            return super().decode(name)

        def encode(self, name):
            """Store file in archive."""
            return super().encode(name)

        @classmethod
        def encode_all(cls, data, outstream):
            """Write all items to archive."""
            format = cls.libarchive_format
            if not format:
                raise ValueError(f'Writing not supported for this format.')
            with libarchive.custom_writer(outstream.write, format) as archive:
                for name, filedict in data.items():
                    payload = filedict.pop('outstream').getvalue()
                    archive.add_file_from_memory(name, len(payload), payload)

        @classmethod
        def decode_all(cls, instream):
            """Read all items from archive."""
            data = {}
            try:
                with libarchive.stream_reader(instream) as archive:
                    for entry in archive:
                        name = Path(str(entry)).as_posix()
                        if name == '.':
                            continue
                        if entry.isdir:
                            data[name + '/'] = b''
                        elif entry.isreg:
                            data[name] = b''.join(list(entry.get_blocks()))
                            for path in Path(name).parents:
                                if path != Path('.'):
                                    data[f'{path}/'] = b''
            except libarchive.ArchiveError as e:
                raise FileFormatError(e) from e
            return data


    @containers.register(
        name='rar',
        patterns=('*.rar',),
        magic=(
            # RAR 1.5 -- 4.0
            b'\x52\x61\x72\x21\x1A\x07\x00',
            # RAR 5+
            b'\x52\x61\x72\x21\x1A\x07\x01\x00',
        ),
    )
    class RARContainer(LibArchiveContainer):
        """RAR archive."""


    @containers.register(
        name='7zip',
        patterns=('*.7z',),
        magic=(
            b'7z\xBC\xAF\x27\x1C',
        ),
    )
    class SevenZipContainer(LibArchiveContainer):
        """7-Zip archive."""
        libarchive_format = '7zip'


    @containers.register(
        name='cabinet',
        patterns=('*.cab',),
        magic=(
            b'MSCF',
        ),
    )
    class CabinetContainer(LibArchiveContainer):
        """Microsoft Cabinet (.cab) file."""


    @containers.register(
        name='iso9660',
        patterns=('*.iso',),
    )
    class ISO9660Container(LibArchiveContainer):
        """ISO-9660 cd-rom image file."""
        libarchive_format = 'iso9660'


    @containers.register(
        name='lharc',
        patterns=('*.lha', '*.lzh'),
    )
    class LHArcContainer(LibArchiveContainer):
        """LHarc archive."""


    @containers.register(
        name='cpio',
        patterns=('*.cpio',),
        magic=(
            b'\x71\xC7',
            b'\xC7\x71',
            b'0707',
        ),
    )
    class CPIOContainer(LibArchiveContainer):
        """CPIO archive."""
        libarchive_format = 'cpio'


    @containers.register(
        name='pax',
        patterns=('*.pax',),
    )
    class PaxContainer(LibArchiveContainer):
        """Pax archive."""
        libarchive_format = 'pax'


    @containers.register(
        name='ar',
        patterns=('*.ar',),
        magic=(b'!<arch>\n',),
    )
    class ArContainer(LibArchiveContainer):
        """Ar archive."""
        libarchive_format = 'ar'


    @containers.register(
        name='xar',
        patterns=('*.xar',),
        magic=(b'xar!',),
    )
    class XArContainer(LibArchiveContainer):
        """XAr archive."""
        libarchive_format = 'xar'


    @containers.register(
        name='warc',
        patterns=('*.warc',),
        magic=(b'WARC/',),
    )
    class WARCContainer(LibArchiveContainer):
        """WARC container."""
        libarchive_format = 'warc'



"""
monobit.encoding.charmaps - character maps

(c) 2020--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
import unicodedata
from pathlib import Path
from html.parser import HTMLParser
from importlib.resources import files
from functools import cached_property, wraps, partial

from ..base.binary import align, int_to_bytes
from ..base import reverse_dict
from ..core.labels import Codepoint, Char, to_label, to_range
from .unicode import is_printable, is_fullwidth, unicode_name
from .base import (
    Encoder, EncoderBuilder, NotFoundError, register_reader, encoding_readers
)
from . import tables


class Unicode(Encoder):
    """Convert between unicode and UTF-32 ordinals."""

    def __init__(self):
        """Unicode converter."""
        super().__init__('unicode')

    @staticmethod
    def char(*labels):
        """Convert codepoint to character."""
        for label in labels:
            codepoint = to_label(label)
            if isinstance(codepoint, bytes):
                # ensure codepoint length is a multiple of 4
                codepoint = codepoint.rjust(align(len(codepoint), 2), b'\0')
                # convert as utf-32 chunks
                chars = tuple(
                    chr(int.from_bytes(codepoint[_start:_start+4], 'big'))
                    for _start in range(0, len(codepoint), 4)
                )
                try:
                    return Char(''.join(chars))
                except ValueError:
                    return Char('')

    @staticmethod
    def codepoint(*labels):
        """Convert character to codepoint."""
        for label in labels:
            char = to_label(label)
            if isinstance(char, str):
                # we used to normalise to NFC here, presumably to reduce multi-codepoint situations
                # but it leads to inconsistency between char and codepoint for canonically equivalent chars
                #char = unicodedata.normalize('NFC', char)
                return Codepoint(b''.join(ord(_c).to_bytes(4, 'big') for _c in char))
        return Codepoint()

    def __repr__(self):
        """Representation."""
        return type(self).__name__ + '()'


class Charmap(Encoder):
    """Convert between unicode and ordinals using stored dictionary."""

    def __init__(self, mapping=None, *, name=''):
        """Create charmap from a dictionary codepoint -> char."""
        super().__init__(name=name)
        if not mapping:
            mapping = {}
            name = ''
        # copy dict
        self._ord2chr = {**mapping}

    @cached_property
    def _chr2ord(self):
        return reverse_dict(self._ord2chr)

    def char(self, *labels):
        """Convert codepoint sequence to character, return empty string if missing."""
        for label in labels:
            codepoint = to_label(label)
            if isinstance(codepoint, bytes):
                try:
                    return Char(self._ord2chr[codepoint])
                except KeyError as e:
                    return Char('')

    def codepoint(self, *labels):
        """Convert character to codepoint sequence, return empty tuple if missing."""
        for label in labels:
            char = to_label(label)
            if isinstance(char, str):
                try:
                    return Codepoint(self._chr2ord[char])
                except KeyError as e:
                    return Codepoint()

    @property
    def mapping(self):
        return {**self._ord2chr}

    def __len__(self):
        """Number of defined codepoints."""
        return len(self._ord2chr)

    def __eq__(self, other):
        """Compare to other Charmap."""
        return isinstance(other, Charmap) and (self._ord2chr == other._ord2chr)

    # charmap operations

    def __sub__(self, other):
        """Return encoding with only characters that differ from right-hand side."""
        return Charmap(
            mapping={_k: _v for _k, _v in self._ord2chr.items() if other.char(_k) != _v},
            name=f'[{self.name}]-[{other.name}]'
        )

    def __or__(self, other):
        """Return encoding overlaid with all characters defined in right-hand side."""
        return Charmap(mapping=self._ord2chr | other._ord2chr, name=f'{self.name}')

    def distance(self, other):
        """Return number of different code points."""
        other_only = set(other._ord2chr) - set(self._ord2chr)
        self_only = set(self._ord2chr) - set(other._ord2chr)
        different = set(
            _k for _k, _v in self._ord2chr.items()
            if _k in other._ord2chr and other.char(_k) != _v
        )
        return len(different) + len(other_only) + len(self_only)

    def subset(self, codepoint_range):
        """Return encoding only for given range of codepoints."""
        codepoint_range = list(to_range(codepoint_range))
        return Charmap(
            mapping={
                _k: _v
                for _k, _v in self._ord2chr.items()
                if (_k in codepoint_range) or (len(_k) == 1 and _k[0] in codepoint_range)
            },
            name=self.name
        )

    def shift(self, by=0x80):
        """
        Increment all codepoints by the given amount.

        by: amount to increment
        """
        return Charmap(
            mapping={
                bytes(Codepoint(int(Codepoint(_k))+by)): _v
                for _k, _v in self._ord2chr.items()
            },
            name=f'shift-{by:x}[{self.name}]'
        )

    # representations

    def chart(self, page=0):
        """Chart of page in charmap."""
        bg = '\u2591'
        cps = range(256)
        cps = (((page, _c) if page else (_c,)) for _c in cps)
        chars = tuple(self.char(_cp) for _cp in cps)
        rows = [(_r, chars[16*_r:16*(_r+1)]) for _r in range(16)]
        # omit empty rows
        for startr, chars in rows:
            if any(chars):
                break
        for stopr, chars in rows[::-1]:
            if any(chars):
                break
        rows = rows[startr:stopr+1]
        def _reprchar(c):
            if c:
                if not is_printable(c):
                    c = '\ufffd'
                if not is_fullwidth(c):
                    c += ' '
                # deal with Nonspacing Marks while keeping table format
                if unicodedata.category(c[:1]) == 'Mn':
                    c = ' ' + c
            else:
                c = bg * 2
            return c
        rows = (
            (_r, bg.join(_reprchar(_c) for _c in _chars))
            for _r, _chars in rows
        )
        return ''.join((
            '#     ', ' '.join(f'_{_c:x}' for _c in range(16)), '\n# ',
            '  +', '-'*48, '-', '\n# ',
            '\n# '.join(
                ''.join((f'{_r:x}_|', bg, _chars, bg))
                for _r, _chars in rows
            )
        ))

    def table(self):
        """Mapping table"""
        return '\n'.join(
            f'0x{_k.hex()}: u+{ord(_v):04X}  # {unicode_name(_v)}' for _k, _v in self._ord2chr.items()
        )

    def __repr__(self):
        """Representation."""
        if self._ord2chr:
            mapping = f'<{len(self._ord2chr)} code points>'
            chart = f'\n{self.chart()}\n'
            return (
                f"{type(self).__name__}(name='{self.name}', mapping={mapping}){chart}"
            )
        return (
            f"{type(self).__name__}()"
        )


class EncoderLoader(EncoderBuilder):
    """Lazily create new encoder from file."""

    def __init__(self, filename, *, format=None, name='', **kwargs):
        if not name:
            name = Path(filename).stem
        filename = str(filename)
        # inputs that look like explicit paths used directly
        # otherwise it's relative to the tables package
        if filename.startswith('/') or filename.startswith('.'):
            path = Path(filename)
        else:
            path = files(tables) / filename
        if not path.exists():
            raise NotFoundError(f'Charmap file `{filename}` does not exist')
        format = format or path.suffix[1:].lower()
        try:
            reader, format_kwargs = encoding_readers[format]
        except KeyError as exc:
            raise NotFoundError(f'Undefined charmap file format {format}.') from exc
        super().__init__(partial(reader, name, path, **format_kwargs, **kwargs))


###############################################################################
# charmap loaders


def _charmap_loader(fn):
    """Decorator for the shared parts of charmap loaders."""

    @wraps(fn)
    def _load(name, path, *args, **kwargs):
        try:
            data = path.read_bytes()
        except EnvironmentError as exc:
            raise NotFoundError(f'Could not load charmap file `{str(path)}`: {exc}')
        if not data:
            raise NotFoundError(f'No data in charmap file `{str(path)}`')
        mapping = fn(data, *args, **kwargs)
        return Charmap(mapping, name=name)

    return _load


@register_reader('txt')
@register_reader('enc')
@register_reader('map')
@register_reader('ucp', separator=':', joiner=',')
@register_reader('adobe', separator='\t', joiner=None, codepoint_column=1, unicode_column=0)
@_charmap_loader
def _from_text_columns(
        data, *, comment='#', separator=None, joiner='+', codepoint_column=0, unicode_column=1,
        codepoint_base=16, unicode_base=16, inline_comments=True, ignore_errors=False,
    ):
    """Extract character mapping from text columns in file data (as bytes)."""
    mapping = {}
    for line in data.decode('utf-8-sig').splitlines():
        # ignore empty lines and comment lines (first char is #)
        if (not line) or (line[0] == comment):
            continue
        if line.startswith('START') or line.startswith('END'):
            # xfonts .enc files - STARTENCODING, STARTMAPPING etc.
            continue
        # strip off comments
        if inline_comments:
            line = line.split(comment)[0]
        # split unicodepoint and hex string
        splitline = line.split(separator)
        if len(splitline) > max(codepoint_column, unicode_column):
            cp_str, uni_str = splitline[codepoint_column], splitline[unicode_column]
            cp_str = cp_str.strip()
            uni_str = uni_str.strip()
            # right-to-left marker in mac codepages
            uni_str = uni_str.replace('<RL>+', '').replace('<LR>+', '')
            # reverse-video marker in kreativekorp codepages
            uni_str = uni_str.replace('<RV>+', '')
            # czyborra's codepages have U+ in front
            if uni_str.upper().startswith('U+'):
                uni_str = uni_str[2:]
            # ibm-ugl codepage has U in front
            if uni_str.upper().startswith('U'):
                uni_str = uni_str[1:]
            # czyborra's codepages have = in front
            if cp_str.upper().startswith('='):
                cp_str = cp_str[1:]
            try:
                # allow sequence of codepoints
                # multibyte code points can also be given as single large number
                # note that the page bytewidth of the codepoints is assumed to be 1
                cp_point = b''.join(
                    int_to_bytes(int(_substr, codepoint_base))
                    for _substr in cp_str.split(joiner)
                )
                if unicode_base == 'char':
                    # the character itself is in the column, utf-8 encoded
                    char = uni_str
                else:
                    # allow sequence of unicode code points separated by 'joiner'
                    char = ''.join(
                        chr(int(_substr, unicode_base))
                        for _substr in uni_str.split(joiner)
                    )
                if char != '\uFFFD':
                    # u+FFFD replacement character is used to mark undefined code points
                    mapping[cp_point] = char
            except (ValueError, TypeError) as e:
                # ignore malformed lines
                if not ignore_errors:
                    logging.warning('Could not parse line in text charmap file: %s [%s]', e, repr(line))
    return mapping


@register_reader('ucm')
@_charmap_loader
def _from_ucm_charmap(data):
    """Extract character mapping from icu ucm / linux charmap file data (as bytes)."""
    # only deals with sbcs
    comment = '#'
    escape = '\\'
    # precision indicator
    precision = '|'
    mapping = {}
    parse = False
    for line in data.decode('utf-8-sig').splitlines():
        # ignore empty lines and comment lines (first char is #)
        if (not line) or (line[0] == comment):
            continue
        if line.startswith('<comment_char>'):
            comment = line.split()[-1].strip()
        elif line.startswith('<escape_char>'):
            escape = line.split()[-1].strip()
        elif line.startswith('CHARMAP'):
            parse = True
            continue
        elif line.startswith('END CHARMAP'):
            parse = False
        if not parse:
            continue
        # split columns
        splitline = line.split()
        # ignore malformed lines
        exc = ''
        cp_bytes, uni_str = '', ''
        for item in splitline:
            if item.startswith('<U'):
                # e.g. <U0000> or <U2913C>
                uni_str = item[2:-1]
            elif item.startswith(escape + 'x'):
                cp_str = item.replace(escape + 'x', '')
                cp_bytes = bytes.fromhex(cp_str)
            elif item.startswith(precision):
                # precision indicator
                # |0 - A “normal”, roundtrip mapping from a Unicode code point and back.
                # |1 - A “fallback” mapping only from Unicode to the codepage, but not back.
                # |2 - A subchar1 mapping. The code point is unmappable, and if a substitution is
                #      performed, then the subchar1 should be used rather than the subchar.
                #      Otherwise, such mappings are ignored.
                # |3 - A “reverse fallback” mapping only from the codepage to Unicode, but not back
                #      to the codepage.
                # |4 - A “good one-way” mapping only from Unicode to the codepage, but not back.
                if item[1:].strip() != '0':
                    # only accept 'normal' mappings
                    # should we also allow "reverse fallback" ?
                    break
        else:
            if not uni_str or not cp_str:
                logging.warning('Could not parse line in ucm charmap file: %s.', repr(line))
                continue
            if cp_bytes in mapping:
                logging.debug('Ignoring redefinition of code point %s', cp_bytes)
            else:
                mapping[cp_bytes] = chr(int(uni_str, 16))
    return mapping


@register_reader('html')
@_charmap_loader
def _from_wikipedia(data, table=0, column=0, range=None):
    """
    Scrape charmap from table in Wikipedia.
    Reads matrix tables with class="chset".
    table: target table; 0 for 1st chset table, etc.
    column: target column if multiple unicode points provided per cell.
    range: range to read, read all if range is empty
    """

    class _WikiParser(HTMLParser):
        """HTMLParser object to read Wikipedia tables."""

        def __init__(self):
            """Set up Wikipedia parser."""
            super().__init__()
            # output dict
            self.mapping = {}
            # state variables
            # parsing chset table
            self.table = False
            self.count = 0
            # data element
            self.td = False
            # the unicode point is surrounded by <small> tags
            self.small = False
            # parse row header
            self.th = False
            # current codepoint
            self.current = 0

        def handle_starttag(self, tag, attrs):
            """Change state upon encountering start tag."""
            attrs = dict(attrs)
            if (
                    tag == 'table'
                    and 'class' in attrs
                    and 'chset' in attrs['class']
                ):
                if self.count == table:
                    self.table = True
                    self.th = False
                    self.td = False
                    self.small = False
                self.count += 1
            elif self.table:
                if tag == 'td':
                    self.td = True
                    self.small = False
                elif tag == 'small':
                    self.small = True
                elif tag == 'th':
                    self.th = True

        def handle_endtag(self, tag):
            """Change state upon encountering end tag."""
            if tag == 'table':
                self.table = False
                self.th = False
                self.td = False
                self.small = False
            elif tag == 'td':
                self.td = False
                self.current += 1
            elif tag == 'style':
                self.small = False
            elif tag == 'th':
                self.th = False

        def handle_data(self, data):
            """Parse cell data, depending on state."""
            # row header provides first code point of the row
            if self.th and len(data) == 2 and data[-1] == '_':
                self.current = int(data[0],16) * 16
            # unicode point in <small> tag in table cell
            if self.td and self.small:
                cols = data.split()
                if len(cols) > column:
                    data = cols[column]
                if len(data) >= 4:
                    # unicode point
                    if data.lower().startswith('u+'):
                        data = data[2:]
                    # pylint: disable=unsupported-membership-test
                    if not range or self.current in range:
                        try:
                            char = chr(int(data, 16))
                        except ValueError:
                            # not a unicode point
                            pass
                        else:
                            self.mapping[bytes((self.current,))] = char

    parser = _WikiParser()
    parser.feed(data.decode('utf-8-sig'))
    return parser.mapping



"""
monobit.storage.wrapperformats.offset - binary files at offset in binary files

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from ..base import decoders
from ..streams import Stream


@decoders.register(
    name='offset',
)
def decode_offset(stream, *, offset:int=0):
    """
    Binary file at an offset in another binary file.

    offset: offset to use, in bytes.
    """
    stream.read(offset)
    # Stream wrapper will drop a seek() anchor
    return Stream(stream, mode=stream.mode, name=stream.name)



"""
monobit.storage.wrapperformats.basic - binary files embedded in BASIC source files

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from pathlib import Path
from io import BytesIO

from ..base import encoders, decoders
from ..streams import Stream, DelayedWriterStream


@decoders.register(
    name='basic',
    patterns=('*.bas',),
)
def decode_basic(instream):
    """
    Extract binary file encoded in DATA lines in classic BASIC source code.
    Tokenised BASIC files are not supported.
    """
    infile = instream.text
    data = []
    for line in infile:
        _, _, dataline = line.partition('DATA')
        dataline = dataline.strip()
        if not dataline:
            continue
        values = dataline.split(',')
        data.append(bytes(_int_from_basic(_s) for _s in values))
    name = Path(instream.name).stem
    return Stream.from_data(b''.join(data), mode='r', name=name)


@encoders.register(linked=decode_basic)
def encode_basic(
        outstream,
        *,
        line_number_start:int=1000,
        line_number_inc:int=10,
        bytes_per_line:int=8
    ):
    """
    Write binary file encoded into DATA lines in classic BASIC source code.
    Tokenised BASIC files are not supported.

    line_number_start: line number of first DATA line (-1 for no line numbers; default: 1000)
    line_number_inc: increment between line numbers (default: 10)
    bytes_per_line: number of encoded bytes in a source line (default: 8)
    """
    encode_func = _do_encode_basic
    name = Path(outstream.name).stem
    return DelayedWriterStream(
        outstream, encode_func, name,
        line_number_start=line_number_start,
        line_number_inc=line_number_inc,
        bytes_per_line=bytes_per_line,
    )


def _do_encode_basic(
        data, outstream,
        *,
        line_number_start,
        line_number_inc,
        bytes_per_line,
    ):
    outfile = outstream.text
    if (
            line_number_inc <= 0
            and line_number_start is not None and line_number_start > -1
        ):
        raise ValueError('line_number_inc must be > 0')
    line_number = None
    if line_number_start is not None and line_number_start >= 0:
        line_number = line_number_start
    with BytesIO(data) as instream:
        while True:
            data = instream.read(bytes_per_line)
            if not data:
                return
            line = ', '.join(_int_to_basic(_b) for _b in data)
            if line_number is not None:
                outfile.write(f'{line_number} ')
                line_number += line_number_inc
            outfile.write(f'DATA {line}\n')


def _int_to_basic(value):
    """Output hex number in BASIC format."""
    return f'&h{value:02x}'


def _int_from_basic(cvalue):
    """Parse integer from BASIC code."""
    cvalue = cvalue.strip().lower()
    if cvalue.startswith('&h'):
        cvalue = '0x' + cvalue[2:]
    return int(cvalue, 0)



"""
monobit.storage.containers - base classes for containers

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
import itertools
from io import BytesIO
from pathlib import Path

from .streams import Stream, KeepOpen


class Container:
    """Base class for multi-stream containers."""

    def __init__(self, mode='r', name=''):
        self.mode = mode[:1]
        self.name = name
        self.refcount = 0
        self.closed = False

    def __repr__(self):
        """String representation."""
        return (
            f"<{type(self).__name__} "
            f"mode='{self.mode}' name='{self.name}'"
            f"{' [closed]' if self.closed else ''}>"
        )

    def __enter__(self):
        # we don't support nesting the same archive
        assert self.refcount == 0
        self.refcount += 1
        logging.debug('Entering %r', self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.refcount -= 1
        if exc_type == BrokenPipeError:
            return True
        logging.debug('Exiting %r', self)
        self.close()

    def close(self):
        """Close the archive."""
        self.closed = True

    def decode(self, name, **kwargs):
        """Open a binary stream to read from the container."""
        raise NotImplementedError

    def encode(self, name, **kwargs):
        """Open a binary stream to write to the container."""
        raise NotImplementedError

    def is_dir(self, name):
        """Item at `name` is a directory."""
        raise NotImplementedError

    def iter_sub(self, prefix):
        """List contents of a subpath."""
        raise NotImplementedError


class Archive(Container):
    """Base class for multi-file archives."""

    def __init__(self, file, mode='r'):
        """Create archive object."""
        super().__init__(mode, file.name)

    @property
    def root(self):
        """Root directory for archive - elided on read, auto-added on write."""
        if not hasattr(self, '_root'):
            # on output, put all files in a directory with the same name as the archive (without suffix)
            stem = Path(self.name).stem
            if self.mode == 'w':
                self._root = stem
            else:
                # on read, only set root if it is a common parent
                self._root = ''
                if all(Path(_item).is_relative_to(stem) for _item in self.list()):
                    self._root = stem
        return self._root

    def iter_sub(self, prefix):
        """List contents of a subpath."""
        subs = list(
            str(Path(_name).relative_to(self.root)) for _name in self.list()
            if Path(_name).parent == Path(self.root) / prefix
        )
        return subs

    def is_dir(self, name):
        """Item at 'name' is a directory."""
        name = Path(self.root) / name
        if Path(name) == Path(self.root):
            return True
        ziplist = self.list()
        # str(Path) does not end in /
        if f'{name}/' in ziplist:
            return True
        if f'{name}' in ziplist:
            return False
        raise FileNotFoundError(
            f"File '{name}' not found in archive {self}."
        )

    def list(self):
        """List full contents of archive."""
        raise NotImplementedError()


class FlatFilterContainer(Archive):
    """Archive implementation based on filter logic."""

    def __init__(self, stream, mode='r'):
        # private fields
        self._wrapped_stream = stream
        self._data = {}
        self._files = {}
        super().__init__(stream, mode)
        self._get_data()

    def close(self):
        """Close the archive, ignoring errors."""
        if self.mode == 'w' and not self.closed:
            self.encode_all(self._files, self._wrapped_stream)
        self._wrapped_stream.close()
        super().close()

    def list(self):
        """List full contents of archive."""
        return (
            tuple(self._data.keys())
            + tuple(self._files.keys())
        )

    def decode(self, name):
        """Open input stream on filter container."""
        filename = str(Path(self.root) / name)
        name = str(name)
        try:
            data = self._data[filename]
        except KeyError:
            if f'{filename}/' in self._data:
                raise IsADirectoryError(f"'{name}' is a directory")
            raise FileNotFoundError(
                f"No file with name '{name}' found in archive."
            )
        return Stream.from_data(data, mode='r', name=name)

    def encode(self, name, **kwargs):
        """Open output stream on filter container."""
        filename = str(Path(self.root) / name)
        name = str(name)
        if filename in self._files:
            raise FileExistsError(
                f"Cannot create multiple files of the same name '{name}'"
            )
        newfile = Stream(KeepOpen(BytesIO()), mode='w', name=name)
        self._files[filename] = {
            'outstream': newfile,
            **kwargs
        }
        return newfile

    def _get_data(self):
        """Read contents of archive into memory."""
        if self.mode == 'w':
            return
        if self._data:
            return
        self._data = self.decode_all(self._wrapped_stream)
        # create directory items, if not present
        for name in list(self._data.keys()):
            if '/' in name:
                for parent in Path(name).parents:
                    if parent != Path('.'):
                        self._data[f'{parent}/'] = b''

    @classmethod
    def decode_all(cls, instream):
        """Decode all files in readable archive."""
        raise NotImplementedError

    @classmethod
    def encode_all(cls, filedict, outstream, **kwargs):
        """Encode all files in writable archive."""
        raise NotImplementedError


class SerialContainer(FlatFilterContainer):

    @classmethod
    def encode_all(cls, data, outstream, **kwargs):
        outstream.write(cls._head())
        for count, (name, filedict) in enumerate(data.items()):
            filedata = filedict.pop('outstream').getvalue()
            cls._encode(
                name, filedata, outstream,
                **filedict
            )
            if count < len(data) - 1:
                outstream.write(cls._separator())
        outstream.write(cls._tail())

    @classmethod
    def _head(cls):
        return b''

    @classmethod
    def _tail(cls):
        return b''

    @classmethod
    def _separator(cls):
        return b''

    @classmethod
    def _encode(cls, name, data, outstream, **kwargs):
        raise NotImplementedError


class SerialTextContainer(SerialContainer):

    @classmethod
    def encode_all(cls, data, outstream, **kwargs):
        outstream = outstream.text
        return super().encode_all(data, outstream, **kwargs)

    @classmethod
    def _head(cls):
        return ''

    @classmethod
    def _tail(cls):
        return ''

    @classmethod
    def _separator(cls):
        return ''



"""
monobit.render.chart - create font chart

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from itertools import product
from pathlib import Path

from monobit.base.binary import ceildiv
from monobit.base import Props, Coord, RGB
from monobit.core import Codepoint, Glyph, Char
from monobit.storage import savers
from monobit.plumbing import scriptable
from monobit.encoding.unicode import is_showable
from monobit.storage.utils.limitations import ensure_single
from .glyphmap import GlyphMap


@savers.register(name='chart')
def save_chart(
        fonts, outstream, *,
        glyphs_per_line:int=16,
        margin:Coord=Coord(0, 0),
        padding:Coord=Coord(1, 1),
        scale:Coord=Coord(1, 1),
        direction:str=None,
        border:str=' ',
        inklevels:tuple[str]=(' ', '@'),
        codepoint_range:tuple[Codepoint]=None,
        grid_positioning:bool=False,
        max_labels:int=1,
    ):
    """
    Export font to text chart.

    glyphs_per_line: number of glyphs per line in glyph chart (default: 16)
    margin: number of pixels in X,Y direction around glyph chart (default: 0x0)
    padding: number of pixels in X,Y direction between glyphs (default: 1x1)
    scale: number of pixels in X,Y direction per glyph bit (default: 1x1)
    direction: two-part string such as 'left-to-right top-to-bottom'. Default: font direction.
    border: character to use for border pixels (default: space)
    inklevels: characters to use for pixels (default: space, '2')
    codepoint_range: range of codepoints to include (includes bounds and undefined codepoints; default: all codepoints)
    grid_positioning: place codepoints on corresponding grid positions, leaving gaps if undefined (default: false)
    max_labels: maximum number of labels to show per glyph (default: 1)
    """
    glyph_map = create_chart(
        fonts,
        glyphs_per_line=glyphs_per_line,
        margin=margin,
        padding=padding,
        scale=scale,
        direction=direction,
        codepoint_range=codepoint_range,
        grid_positioning=grid_positioning,
        max_labels=max_labels,
    )
    outstream.text.write(
        glyph_map.as_text(border=border, inklevels=inklevels)
    )


@savers.register(name='blocks')
def save_blocks(
        fonts, outstream, *,
        glyphs_per_line:int=16,
        margin:Coord=Coord(0, 0),
        padding:Coord=Coord(1, 1),
        scale:Coord=Coord(1, 1),
        direction:str=None,
        resolution:Coord=Coord(1, 1),
        codepoint_range:tuple[Codepoint]=None,
        grid_positioning:bool=False,
        max_labels:int=1,
    ):
    """
    Export font to blocks chart.

    glyphs_per_line: number of glyphs per line in glyph chart (default: 16)
    margin: number of pixels in X,Y direction around glyph chart (default: 0x0)
    padding: number of pixels in X,Y direction between glyphs (default: 1x1)
    scale: number of pixels in X,Y direction per glyph bit (default: 1x1)
    direction: two-part string such as 'left-to-right top-to-bottom'. Default: font direction.
    resolution: blocks per text character; 1x1 (default), 1x2, 1x3, 1x4, 2x1, 2x3, 2x4
    codepoint_range: range of codepoints to include (includes bounds; default: all codepoints)
    grid_positioning: place codepoints on corresponding grid positions, leaving gaps if undefined (default: false)
    max_labels: maximum number of labels to show per glyph (default: 1)
    """
    glyph_map = create_chart(
        fonts,
        glyphs_per_line=glyphs_per_line,
        margin=margin,
        padding=padding,
        scale=scale,
        direction=direction,
        codepoint_range=codepoint_range,
        max_labels=max_labels,
        grid_positioning=grid_positioning,
        label_height=resolution.y,
    )
    outstream.text.write(glyph_map.as_blocks(resolution=resolution))


@savers.register(name='shades')
def save_shades(
        fonts, outstream, *,
        glyphs_per_line:int=16,
        margin:Coord=Coord(0, 0),
        padding:Coord=Coord(1, 1),
        scale:Coord=Coord(1, 1),
        direction:str=None,
        border:RGB=None,
        paper:RGB=RGB(0, 0, 0),
        ink:RGB=RGB(255, 255, 255),
        codepoint_range:tuple[Codepoint]=None,
        grid_positioning:bool=False,
        max_labels:int=1,
    ):
    """
    Export font to ansi-coloured blocks chart.

    glyphs_per_line: number of glyphs per line in glyph chart (default: 16)
    margin: number of pixels in X,Y direction around glyph chart (default: 0x0)
    padding: number of pixels in X,Y direction between glyphs (default: 1x1)
    scale: number of pixels in X,Y direction per glyph bit (default: 1x1)
    direction: two-part string such as 'left-to-right top-to-bottom'. Default: font direction.
    paper: background colour R,G,B 0--255 (default: 0,0,0)
    ink: full-intensity foreground colour R,G,B 0--255 (default: 255,255,255)
    border: border colour R,G,B 0--255 (default: terminal background)
    codepoint_range: range of codepoints to include (includes bounds; default: all codepoints)
    grid_positioning: place codepoints on corresponding grid positions, leaving gaps if undefined (default: false)
    max_labels: maximum number of labels to show per glyph (default: 1)
    """
    glyph_map = create_chart(
        fonts,
        glyphs_per_line=glyphs_per_line,
        margin=margin,
        padding=padding,
        scale=scale,
        direction=direction,
        codepoint_range=codepoint_range,
        max_labels=max_labels,
        grid_positioning=grid_positioning,
    )
    outstream.text.write(glyph_map.as_shades(paper=paper, border=border, ink=ink))



def create_chart(
        fonts, *,
        glyphs_per_line,
        margin,
        padding,
        scale,
        direction,
        codepoint_range,
        lines_per_page=None,
        max_labels=0,
        label_height=1,
        grid_positioning=False,
    ):
    """Create chart glyph map of font."""
    font = ensure_single(fonts)
    font = font.equalise_horizontal()
    if grid_positioning:
        font = grid_resample(font, glyphs_per_line, codepoint_range)
    elif codepoint_range:
        font = font.subset(codepoint_range)
    font = font.stretch(*scale)
    # create extra padding space to allow for labels
    if max_labels:
        label_padding = (max_labels+1) * label_height
    else:
        label_padding = 0
    padding = Coord(padding.x, padding.y + label_padding)
    margin = Coord(margin.x, margin.y + label_padding)
    glyph_map = grid_map(
        font,
        glyphs_per_line=glyphs_per_line,
        lines_per_page=lines_per_page,
        margin=margin, padding=padding,
        direction=direction,
    )
    right_align = aligns_right(direction)
    for entry in glyph_map:
        for count, label in enumerate(entry.glyph.get_labels()):
            if count >= max_labels:
                break
            if right_align:
                x = entry.x + entry.glyph.width
            else:
                x = entry.x
            glyph_map.append_label(
                format_label(label),
                x,
                label_height * (ceildiv(entry.y + entry.glyph.height, label_height) + (count + 1)),
                sheet=entry.sheet,
                right_align=right_align,
            )
    return glyph_map


def aligns_right(direction):
    """Determine if a given direction is right-aligning."""
    if not direction:
        return False
    dir_0, _, dir_1 = direction.lower().partition(' ')
    return dir_0[:1] == 'r' or dir_1[:1] == 'r'


def format_label(label):
    """Format glyph label for charts."""
    if isinstance(label, Char):
        ucstr = ', '.join(f'u+{ord(_uc):04x}' for _uc in label.value)
        if all(is_showable(_uc) for _uc in label.value):
            return f'{ucstr} {label.value}'
        return ucstr
    return str(label)


###############################################################################
# grid functions

def grid_resample(font, glyphs_per_line, codepoint_range):
    """Resample font for grid representation."""
    if not codepoint_range:
        try:
            codepoint_range = range(
                # start at a codepoint that is a multiple of the number of columns
                glyphs_per_line * (int(min(font.get_codepoints())) // glyphs_per_line),
                int(max(font.get_codepoints()))+1
            )
        except ValueError:
            # empty sequence
            raise ValueError('No codepoint labels found.')
    font = font.resample(codepoint_range, missing='empty', relabel=False)
    return font


def grid_map(
        font, *,
        # glyphs_per_line chooses rows/cols depending on render direction
        # set to None or 0 to mean 'as many as needed'
        glyphs_per_line=None,
        lines_per_page=None,
        margin=(0, 0), padding=(0, 0),
        direction=None,
        invert_y=False,
    ):
    """
    Create glyph grid(s) for font charts.
    """
    padding = Coord(*padding)
    margin = Coord(*margin)
    # work out image geometry
    step_x = font.raster_size.x + padding.x
    step_y = font.raster_size.y + padding.y
    direction = direction or font.direction
    if direction[:1].lower() in ('t', 'b'):
        rows, columns = glyphs_per_line, lines_per_page
    else:
        columns, rows = glyphs_per_line, lines_per_page
    # at most one of these may be None or 0
    rows = rows or ceildiv(len(font.glyphs), columns)
    columns = columns or ceildiv(len(font.glyphs), rows)
    glyphs_per_page = rows * columns
    glyph_pages = tuple(
        font.glyphs[_s : _s + glyphs_per_page]
        for _s in range(0, len(font.glyphs), glyphs_per_page)
    )
    # horizontal alignment (left or right)
    # note that we have equalised glyphs to the same height
    # so vertical alignment is not needed
    right_align = aligns_right(direction)
    # output glyph maps
    glyph_map = GlyphMap(
        Props(
            glyph=_glyph, sheet=_sheet,
            x=(
                margin.x + col*step_x
                + (font.raster_size.x - _glyph.width if right_align else 0)
            ),
            y=margin.y + row*step_y,
        )
        for _sheet, _glyph_page in enumerate(glyph_pages)
        for _glyph, (row, col) in zip(
            _glyph_page,
            grid_traverser(columns, rows, direction, invert_y)
        )
    )
    # use blank glyphs for grid bounds
    glyph_map.append_glyph(Glyph(), 0, 0, sheet=0)
    glyph_map.append_glyph(
        Glyph(),
        2 * margin.x + columns*step_x - padding.x,
        2 * margin.y + rows*step_y - padding.y,
        sheet=0
    )
    return glyph_map


def grid_traverser(columns, rows, direction, invert_y=False):
    """
    Traverse a glyph chart in the specified order and directions.

    direction: a pair of characters of the form 'a b'
               where a, b can be 'l', 'r', 't', 'b'
    invert_y: positive is down
    """
    glyph_dir, _, line_dir = direction.lower().partition(' ')
    glyph_dir = glyph_dir[:1] or 'l'
    line_dir = line_dir[:1] or 't'
    vertical = glyph_dir in ('t', 'b')
    if vertical:
        dir_y, dir_x = glyph_dir, line_dir
    else:
        dir_x, dir_y = glyph_dir, line_dir
    if dir_x == 'l':
        x_traverse = range(columns)
    else:
        x_traverse = range(columns-1, -1, -1)
    if dir_y == 't':
        y_traverse = range(rows-1, -1, -1)
    else:
        y_traverse = range(rows)
    if invert_y:
        y_traverse = reversed(y_traverse)
    if vertical:
        return (reversed(_p) for _p in product(x_traverse, y_traverse))
    else:
        return product(y_traverse, x_traverse)



"""
monobit.storage.fontformats.wsfont - NetBSD wsfont binaries and C headers

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from io import BytesIO

from monobit.storage import loaders, savers
from monobit.core import Glyph, Font
from monobit.base import Props, reverse_dict, FileFormatError
from monobit.base.struct import little_endian as le
from monobit.base.binary import ceildiv
from monobit.encoding import EncodingName

from monobit.storage.utils.source import CCodeReader, CCodeWriter
from monobit.storage.utils.limitations import ensure_single, make_contiguous, ensure_charcell

from .raw import load_bitmap, save_bitmap


# https://ftp.iij.ad.jp/pub/NetBSD/NetBSD-release-10/xsrc/local/programs/bdfload/bdfload.c

_WSF_MAGIC = b'WSFT'

_WSF_HEADER = le.Struct(
    magic='4s',
    name='64s',
    firstchar='uint32',
    numchars='uint32',
    encoding='uint32',
    fontwidth='uint32',
    fontheight='uint32',
    # should be strike width given in bytes
    # but in NetBSD kernel header files this is *sometimes* the pixel width
    # so the field may well be ignored by the kernel
    stride='uint32',
    bitorder='uint32',
    byteorder='uint32',
)

#define WSDISPLAY_FONTORDER_KNOWN 0     /* i.e, no need to convert */
#define WSDISPLAY_FONTORDER_L2R 1
#define WSDISPLAY_FONTORDER_R2L 2
_WSF_ORDER = {
    0: 'left',
    1: 'left',
    2: 'right',
}


# > ISO-8859-1 encoding
#define WSDISPLAY_FONTENC_ISO 0
# > IBM CP437 encoding
#define WSDISPLAY_FONTENC_IBM 1
# > the custom encoding of the supplemental
# > fonts which came with the BSD ``pcvt'' console driver
#define WSDISPLAY_FONTENC_PCVT 2
# > ISO-8859-7 (Greek) encoding
#define WSDISPLAY_FONTENC_ISO7 3 /* greek */
# > ISO-8859-2 (Eastern European) encoding
#define WSDISPLAY_FONTENC_ISO2 4 /* east european */
# > KOI8-R (Russian) encoding
#define WSDISPLAY_FONTENC_KOI8_R 5 /* russian */
_WSF_ENCODING = {
    0: EncodingName('iso-8859-1'),
    1: EncodingName('cp437'),
    2: None,
    3: EncodingName('iso-8859-7'),
    4: EncodingName('iso-8859-2'),
    5: EncodingName('koi8-r'),
}
_TO_WSF_ENCODING = reverse_dict(_WSF_ENCODING)



_KEY_TYPE = 'wsdisplay_font'

_HEADER_FIELDS = (
    'name', 'firstchar', 'numchars', 'encoding', 'fontwidth', 'fontheight', 'stride', 'bitorder', 'byteorder', 'data'
)

# constants from # https://ftp.iij.ad.jp/pub/NetBSD/NetBSD-release-10/xsrc/local/programs/bdfload/bdfload.c

_ORDER_CONST = {
    # /* i.e, no need to convert */
    # whatever that means
    'WSDISPLAY_FONTORDER_KNOWN': 0,
    'WSDISPLAY_FONTORDER_L2R': 1,
    'WSDISPLAY_FONTORDER_R2L': 2,
}

_ENCODING_CONST = {
    # > ISO-8859-1 encoding
    'WSDISPLAY_FONTENC_ISO': 0,
    # > IBM CP437 encoding
    'WSDISPLAY_FONTENC_IBM': 1,
    # > the custom encoding of the supplemental
    # > fonts which came with the BSD ``pcvt'' console driver
    'WSDISPLAY_FONTENC_PCVT': 2,
    # > ISO-8859-7 (Greek) encoding
    'WSDISPLAY_FONTENC_ISO7': 3,
    # > ISO-8859-2 (Eastern European) encoding
    'WSDISPLAY_FONTENC_ISO2': 4,
    # > KOI8-R (Russian) encoding
    'WSDISPLAY_FONTENC_KOI8_R': 5,
}


@loaders.register(
    name='netbsd',
    patterns=('*.h',),
)
def load_netbsd(instream):
    """Load font from NetBSD wsfont C header."""
    instream = instream.text
    identifier = ''
    data = {}
    headers = []
    for line in instream:
        line = CCodeReader.strip_line_comments(line, instream)
        if CCodeReader.assign in line:
            identifier, _, _ = line.partition(CCodeReader.assign)
            logging.debug('Found assignment to `%s`', identifier)
            if _KEY_TYPE in identifier:
                headers.append(_read_header(line, instream))
            elif CCodeReader.delimiters[0] in line:
                identifier = CCodeReader.clean_identifier(identifier)
                coded_data = CCodeReader.read_array(instream, line)
                data[identifier] = CCodeReader.decode_array(coded_data)
    return tuple(
        load_wsfont_bitmap(BytesIO(data[_header.data]), _header)
        for _header in headers
    )


def _read_header(line, instream):
    """Read the wsfont header from a c file."""
    headerlist = CCodeReader.read_array(instream, line)
    header = CCodeReader.decode_struct(headerlist, _HEADER_FIELDS)
    header.name = header.name.strip('"').encode('ascii', 'replace')
    header.firstchar = CCodeReader.decode_int(header.firstchar)
    # ignore numchars - it is often an expression
    # and we can work it out anyway
    header.numchars = None
    header.fontwidth = CCodeReader.decode_int(header.fontwidth)
    header.fontheight = CCodeReader.decode_int(header.fontheight)
    header.stride = CCodeReader.decode_int(header.stride)
    try:
        header.encoding = _ENCODING_CONST[header.encoding]
    except KeyError:
        header.encoding = CCodeReader.decode_int(header.encoding)
    try:
        header.bitorder = _ORDER_CONST[header.bitorder]
    except KeyError:
        header.bitorder = CCodeReader.decode_int(header.bitorder)
    try:
        header.byteorder = _ORDER_CONST[header.byteorder]
    except KeyError:
        header.byteorder = CCodeReader.decode_int(header.byteorder)
    return header


@savers.register(linked=load_netbsd)
def save_netbsd(
        fonts, outstream, *,
        byte_order:str=None,
        bit_order:str='left',
        add_visuals:bool=True,
    ):
    """
    Save to a NetBSD wsfont header.

    bit_order: 'left' for most significant bit left (default), or 'right'
    byte_order: 'left' or 'right'; default: same as bit-order
    add_visuals: add comments with visual representation of glyphs (default: True)
    """
    for font in fonts:
        _write_netbsd(font, outstream, byte_order, bit_order, add_visuals)


def _write_netbsd(font, outstream, byte_order, bit_order, add_visuals):
    """Write single NetBSD wsfont header."""
    identifier = CCodeWriter.to_identifier(font.name)
    header, data = convert_to_wsfont(font, byte_order, bit_order)
    header = Props(**vars(header))
    header.name = '"' + header.name.decode('ascii', 'replace') + '"'
    header.data = f'{identifier}_data'
    header.encoding = reverse_dict(_ENCODING_CONST)[header.encoding]
    header.bitorder = reverse_dict(_ORDER_CONST)[header.bitorder]
    header.byteorder = reverse_dict(_ORDER_CONST)[header.byteorder]
    headerstr = CCodeWriter.encode_struct(header, _HEADER_FIELDS)
    if add_visuals:
        arraystr = CCodeWriter.encode_array(
            data, bytes_per_line=header.stride,
            lines_per_block=header.fontheight,
            add_visuals=True, index=header.firstchar,
        )
    else:
        # compact format
        arraystr = CCodeWriter.encode_array(
            data, bytes_per_line=header.stride * header.fontheight,
        )
    outstream = outstream.text
    outstream.write('\n\n')
    if font.get_comment():
        outstream.write(f'/*\n{font.get_comment()}\n*/\n\n')
    outstream.write(f'static u_char {header.data}[];\n\n')
    outstream.write(f'struct wsdisplay_font {identifier} = {headerstr};\n\n')
    outstream.write(f'static u_char {header.data}[] = {arraystr};\n')


###############################################################################

@loaders.register(
    name='wsfont',
    magic=(_WSF_MAGIC,),
    patterns=('*.wsf',),
)
def load_wsfont(instream):
    """Load a wsfont .wsf font."""
    header = _WSF_HEADER.read_from(instream)
    if header.magic != _WSF_MAGIC:
        raise FileFormatError(
            f'Not a .wsf file: incorrect signature {header.magic}'
        )
    return load_wsfont_bitmap(instream, header)


def load_wsfont_bitmap(instream, header):
    """Load bitmap font with geometry defined by wsfont header."""
    # if header.stride == header.fontwidth:
    #     header.stride = ceildiv(header.fontwidth, 8)
    font = load_bitmap(
        instream,
        width=header.fontwidth, height=header.fontheight,
        count=header.numchars,
        align=_WSF_ORDER[header.bitorder],
        strike_bytes=header.stride,
        strike_count=1,
        first_codepoint=header.firstchar,
        msb=_WSF_ORDER[header.bitorder],
        byte_swap=(
            header.stride
            if header.byteorder and header.byteorder != header.bitorder
            else 0
        ),
    )
    encoding = _WSF_ENCODING.get(header.encoding, '')
    if encoding:
        font = font.label(char_from=encoding)
    font = font.modify(name=header.name.decode('ascii', 'replace'))
    return font


@savers.register(linked=load_wsfont)
def save_wsfont(
        fonts, outstream, *,
        byte_order:str=None,
        bit_order:str='left',
    ):
    """
    Save to a wsfont .wsf font.

    bit_order: 'left' for most significant bit left (default), or 'right'
    byte_order: 'left' or 'right'; default: same as bit-order
    """
    font = ensure_single(fonts)
    header, data = convert_to_wsfont(font, byte_order, bit_order)
    outstream.write(bytes(header))
    outstream.write(data)


def convert_to_wsfont(font, byte_order, bit_order):
    if not byte_order:
        byte_order = bit_order
    font = ensure_charcell(font)
    if font.encoding not in _TO_WSF_ENCODING and font.get_chars():
        # standard encoding is latin-1 in this format
        font = font.label(codepoint_from='latin-1', overwrite=True)
    font = make_contiguous(font, missing='space')
    codepoints = font.get_codepoints()
    if not codepoints:
        raise ValueError('No storable codepoints found in font.')
    header = _WSF_HEADER(
        magic=_WSF_MAGIC,
        name=font.name.encode('ascii', 'replace'),
        firstchar=min(int(_cp) for _cp in codepoints),
        numchars=len(codepoints),
        encoding=_TO_WSF_ENCODING.get(font.encoding, 0),
        fontwidth=font.cell_size.x,
        fontheight=font.cell_size.y,
        stride=ceildiv(font.cell_size.x, 8),
        bitorder=2 if bit_order.startswith('r') else 1,
        byteorder=2 if byte_order.startswith('r') else 1,
    )
    stream = BytesIO()
    save_bitmap(
        stream, font,
        align=bit_order, msb=bit_order,
        byte_swap=(header.stride if byte_order != bit_order else 0),
    )
    return header, stream.getvalue()



"""
monobit.storage.fontformats.vector.gimms - GIMMS.BIN

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

# GIMMS vector fonts, as decoded by EDWIN
# see https://gtoal.com/vectrex/vector_fonts/gimms/
#     https://gtoal.com/vectrex/vector_fonts/gimms/GIMMS.imp.html

import logging

from monobit.storage import loaders, savers, Magic
from monobit.base import FileFormatError
from monobit.base.struct import StructError, big_endian as be
from monobit.core import Font, Glyph, StrokePath, StrokeMove


_GIMMS_MAGIC = Magic.offset(4) + b'GIMM'

@loaders.register(
    name='gimms',
    magic=(_GIMMS_MAGIC,),
)
def load_gimms(instream):
    """Read GIMMS.BIN font file."""
    fonts = []
    while True:
        font = _read_gimms_resource(instream)
        if not font:
            break
        fonts.append(font)
    return fonts


_HEADER = be.Struct(
    fontlength='uint32',
    gimm='4s',
    number='4s',
    padding='16s',
    scale='uint16',
    xx='uint8',
    yy='uint8',
    start=be.uint16 * 128,
)

_PENUP = 1 << 15
_LASTVECTOR = 1 << 7

_CHARHEADER = be.Struct(
    xbias='uint8',
    xmax='uint8',
    ybias='uint8',
    ymax='uint8',
    # data='uint16',
)


def _read_gimms_resource(instream):
    """Read font resource from GIMMS.BIN file."""
    try:
        gimms_header = _HEADER.read_from(instream)
    except StructError:
        return None
    if gimms_header.gimm != b'GIMM':
        raise FileFormatError(
            f'Not a GIMMS font: `GIMM` signature not found'
        )
    chdata = instream.read(gimms_header.fontlength - _HEADER.size)
    glyphs = []
    for cp, start in enumerate(gimms_header.start):
        if not start:
            continue
        offset = start - 256
        charheader = _CHARHEADER.from_bytes(chdata[offset:])
        offset += _CHARHEADER.size
        # biases will define left_bearing and shift_up
        lastx, lasty = charheader.xbias, charheader.ybias
        path = []
        while True:
            codeword = int(be.uint16.from_bytes(chdata[offset:offset+2]))
            offset += 2
            x = (codeword>>8) & 127
            y = codeword & 127
            if codeword & _PENUP:
                path.append((StrokePath.MOVE, x-lastx, y-lasty))
            else:
                path.append((StrokePath.LINE, x-lastx, y-lasty))
            lastx, lasty = x, y
            if codeword & _LASTVECTOR:
                break
        glyphs.append(
            Glyph.from_path(path, codepoint=cp)
        )
    return Font(glyphs, name=f"GIMMS-{gimms_header.number.decode('latin-1')}")



"""
monobit.storage.fontformats.apple.nfnt - Mac FONT/NFNT fonts

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT

NFNT writer based on the Apple IIgs writer
(c) 2023 Kelvin Sherlock
licence: https://opensource.org/licenses/MIT
"""

import logging
from itertools import chain, accumulate

from monobit.base.binary import bytes_to_bits
from monobit.base.struct import bitfield, big_endian as be, little_endian as le
from monobit.core import Font, Glyph, KernTable, Char, Raster
from monobit.storage import loaders, savers
from monobit.encoding import EncodingName
from monobit.base import Props, NOT_SET, UnsupportedError

from .fond import fixed_to_float
from ..common import MAC_ENCODING


@loaders.register(
    name='nfnt',
    # \x90\0 is not a formal signature, but the most common set of FONT_TYPE flags
    # the \x80 sigs are LISA compressed NFNTs
    magic=(b'\x90\0', b'\xb0\0', b'\x90\x80', b'\xb0\x80'),
    patterns=('*.f',),
)
def load_nfnt(instream, offset:int=0):
    """
    Load font from a bare FONT/NFNT resource.

    offset: starting offset in bytes of the NFNT record in the file (default 0)
    """
    instream.seek(offset)
    data = instream.read()
    fontdata = extract_nfnt(data, 0)
    return convert_nfnt({}, **fontdata)


@savers.register(linked=load_nfnt)
def save_nfnt(
        fonts, outstream,
        create_width_table:bool=True,
        create_height_table:bool=False,
        resample_encoding:EncodingName=NOT_SET,
    ):
    """
    Write font to a bare FONT/NFNT resource.

    create_width_table: include a fractional glyph-width table in the resource (default: True)
    create_height_table: include an image-height table in the resource (default: False)
    resample_encoding: encoding to use for NFNT resources. Must be one of the `mac-` encodings. Default: use font's encoding.
    """
    if len(fonts) > 1:
        logging.warning('NFNT resource can only store one font.')
    font = fonts[0]
    data, _, _ = create_nfnt(
        font, endian='big', ndescent_is_high=True,
        create_width_table=create_width_table,
        create_height_table=create_height_table,
        resample_encoding=resample_encoding,
    )
    outstream.write(data)


##############################################################################

# normalised names so we can test with `in`
MAC_ENCODING_NAMES = tuple(str(EncodingName(_name) for _name in MAC_ENCODING))

# tag used for the 'missing' glyph
MISSING_TAG = 'missing'


##############################################################################
# NFNT/FONT resource

# the Font Type Element
# https://developer.apple.com/library/archive/documentation/mac/Text/Text-251.html#MARKER-9-442
def font_type_struct(base):
    return base.Struct(
        # 15    Reserved. Should be set to 0.
        reserved_15=bitfield('uint16', 1),
        # 14    This bit is set to 1 if the font is not to be expanded to match the screen depth. The
        #       font is for color Macintosh computers only if this bit is set to 1. This is for some
        #       fonts, such as Kanji, which are too large for synthetic fonts to be effective or
        #       meaningful, or bitmapped fonts that are larger than 50 points.
        dont_expand_to_match_screen_depth=bitfield('uint16', 1),
        # 13    This bit is set to 1 if the font describes a fixed-width font, and is set to 0 if the
        #       font describes a proportional font. The Font Manager does not check the setting of this bit.
        fixed_width=bitfield('uint16', 1),
        # 12    Reserved. Should be set to 1.
        reserved_12=bitfield('uint16', 1),
        # 10-11 Reserved. Should be set to 0.
        reserved_10_11=bitfield('uint16', 2),
        # 9     This bit is set to 1 if the font contains colors other than black. This font is for
        #       color Macintosh computers only if this bit is set to 1.
        has_colors=bitfield('uint16', 1),
        # 8     This bit is set to 1 if the font is a synthetic font, created dynamically from the
        #       available font resources in response to a certain color and screen depth combination.
        #       The font is for color Macintosh computers only if this bit is set to 1.
        synthetic=bitfield('uint16', 1),
        # 7     This bit is set to 1 if the font has a font color table ('fctb') resource. The font
        #       is for color Macintosh computers only if this bit is set to 1.
        has_fctb=bitfield('uint16', 1),
        # 4-6   Reserved. Should be set to 0.
        reserved_4_6=bitfield('uint16', 3),
        # 2-3   These two bits define the depth of the font. Each of the four possible values indicates
        #       the number of bits (and therefore, the number of colors) used to represent each pixel
        #       in the glyph images.
        #       Value    Font depth    Number of colors
        #           0    1-bit    1
        #           1    2-bit    4
        #           2    4-bit    16
        #           3    8-bit    256
        #       Normally the font depth is 0 and the glyphs are specified as monochrome images. If
        #       bit 7 of this field is set to 1, a resource of type 'fctb' with the same ID as the font
        #       can optionally be provided to assign RGB colors to specific pixel values.
        #
        # If this font resource is a member of a font family, the settings of bits 8 and 9 of the
        # fontStyle field in this font's association table entry should be the same as the settings of
        # bits 2 and 3 in the fontType field. For more information, see "The Font Association Table"
        # on page 4-89.
        depth=bitfield('uint16', 2),
        # 1    This bit is set to 1 if the font resource contains a glyph-width table.
        has_width_table=bitfield('uint16', 1),
        # 0 This bit is set to 1 if the font resource contains an image height table.
        has_height_table=bitfield('uint16', 1),
    )

# the header of the NFNT is a FontRec
# https://developer.apple.com/library/archive/documentation/mac/Text/Text-214.html
def nfnt_header_struct(base):
    return base.Struct(
        #    {font type}
        fontType=font_type_struct(base),
        #    {character code of first glyph}
        firstChar='uint16',
        #    {character code of last glyph}
        lastChar='uint16',
        #    {maximum glyph width}
        widMax='uint16',
        #    {maximum glyph kern}
        kernMax='int16',
        #    {negative of descent}
        nDescent='int16',
        #    {width of font rectangle}
        fRectWidth='uint16',
        #    {height of font rectangle}
        fRectHeight='uint16',
        #    {offset to width/offset table}
        owTLoc='uint16',
        #    {maximum ascent measurement}
        ascent='uint16',
        #    {maximum descent measurement}
        descent='uint16',
        #    {leading measurement}
        leading='uint16',
        #    {row width of bit image in 16-bit wds}
        rowWords='uint16',
        # followed by:
        # bit image table
        # bitmap location table
        # width offset table
        # glyph-width table
        # image height table
    )


# location table entry
def loc_entry_struct(base):
    return base.Struct(
        offset='uint16',
    )


# width/offset table entry
# Width/offset table. For every glyph in the font, this table contains a word with the glyph offset
# in the high-order byte and the glyph's width, in integer form, in the low-order byte. The value of
# the offset, when added to the maximum kerning  value for the font, determines the horizontal
# distance from the glyph origin to the left edge of the bit image of the glyph, in pixels. If this
# sum is negative, the glyph origin  is to the right of the glyph image's left edge, meaning the
# glyph kerns to the left.  If the sum is positive, the origin is to the left of the image's left
# edge. If the sum equals zero, the glyph origin corresponds with the left edge of the bit image.
# Missing glyphs are represented by a word value of -1. The last word of this table is also -1,
# representing the end.
def wo_entry_struct(base):
    if base == be:
        return base.Struct(
            offset='uint8',
            width='uint8',
        )
    else:
        return base.Struct(
            width='uint8',
            offset='uint8',
        )

# glyph width table entry
# > Glyph-width table. For every glyph in the font, this table contains a word
# > that specifies the glyph's fixed-point glyph width at the given point size
# > and font style, in pixels. The Font Manager gives precedence to the values
# > in this table over those in the font family glyph-width table. There is an
# > unsigned integer in the high-order byte and a fractional part in the
# > low-order byte. This table is optional.
def width_entry_struct(base):
    return base.Struct(
        # divide by 256
        width='uint16',
    )

# height table entry
# Image height table. For every glyph in the font, this table contains a word that specifies the
# image height of the glyph, in pixels. The image height is the height of the glyph image and is
# less than or equal to the font height. QuickDraw uses the image height for improved character
# plotting, because it only draws the visible part of the glyph. The high-order byte of the word is
# the offset from the top of the font rectangle of the first non-blank (or nonwhite) row in the
# glyph, and the low-order byte is the number of rows that must be drawn. The Font Manager creates
# this table.
def height_entry_struct(base):
    return base.Struct(
        offset='uint8',
        height='uint8',
    )


def extract_nfnt(data, offset, endian='big', owt_loc_high=0, font_type=None):
    """Read a MacOS NFNT or FONT resource."""
    # create struct types; IIgs NFNTs are little-endian
    base = {'b': be, 'l': le}[endian[:1].lower()]
    NFNTHeader = nfnt_header_struct(base)
    LocEntry = loc_entry_struct(base)
    WOEntry = wo_entry_struct(base)
    WidthEntry = width_entry_struct(base)
    HeightEntry = height_entry_struct(base)
    # font type override (for IIgs)
    if font_type is not None:
        data = font_type + data[offset+2:]
        offset = 0
    # this is not in the header documentation but is is mentioned here:
    # https://www.kreativekorp.com/swdownload/lisa/AppleLisaFontFormat.pdf
    compressed = endian == 'big' and data[offset+1] & 0x80
    if compressed:
        logging.debug('Uncompressing NFNT resource')
        data = _uncompress_nfnt(data, offset)
        offset = 0
    fontrec = NFNTHeader.from_bytes(data, offset)
    logging.debug('FONT/NFNT header: %s', fontrec)
    if not (fontrec.rowWords and fontrec.widMax and fontrec.fRectWidth and fontrec.fRectHeight):
        logging.debug('Empty FONT/NFNT resource.')
        return dict(glyphs=(), fontrec=fontrec)
    if fontrec.fontType.has_fctb:
        raise UnsupportedError('Colour fonts not supported.')
    # read char tables & bitmaps
    # table offsets
    strike_offset = offset + NFNTHeader.size
    loc_offset = offset + NFNTHeader.size + fontrec.fRectHeight * fontrec.rowWords * 2
    # bitmap strike
    strike = data[strike_offset:loc_offset]
    # location table
    # number of chars: coded chars plus missing symbol
    n_chars = fontrec.lastChar - fontrec.firstChar + 2
    # loc table should have one extra entry to be able to determine widths
    loc_table = LocEntry.array(n_chars+1).from_bytes(data, loc_offset)
    # width offset table
    # the high word of the table's offset (in words) is either:
    # - stored in a separate header (for IIgs)
    # - provided in the repurposed positive nDescent field (dfont)
    # https://developer.apple.com/library/archive/documentation/mac/Text/Text-252.html
    if fontrec.nDescent > 0:
        owt_loc_high = fontrec.nDescent
    # owtTLoc is offset "from itself" to table
    wo_offset = offset + 16 + (fontrec.owTLoc + (owt_loc_high << 16)) * 2
    wo_table = WOEntry.array(n_chars).from_bytes(data, wo_offset)
    # the width-offset table has an extra word:
    # > The last word of this table is also -1, representing the end.
    width_offset = wo_offset + WOEntry.size * (n_chars+1)
    height_offset = width_offset
    # scalable width table
    if fontrec.fontType.has_width_table:
        width_table = WidthEntry.array(n_chars).from_bytes(data, width_offset)
        height_offset += WidthEntry.size * n_chars
    # image height table: this can be deduced from the bitmaps
    # https://developer.apple.com/library/archive/documentation/mac/Text/Text-250.html#MARKER-9-414
    # > The Font Manager creates this table.
    if fontrec.fontType.has_height_table:
        height_table = HeightEntry.array(n_chars).from_bytes(data, height_offset)
    # parse bitmap strike
    bits_per_pixel = 2**fontrec.fontType.depth
    bitmap_strike = Raster.from_bytes(
        strike, stride=fontrec.rowWords*16//bits_per_pixel,
        bits_per_pixel=bits_per_pixel
    )
    # if the font was compressed, we need to XOR the bitmap rows
    # compressed strikes are LISA-only - no need to support greyscale
    if compressed:
        rows = bitmap_strike.as_matrix(inklevels=(False, True))
        xoredrows = rows
        rows = [xoredrows[0]]
        for row in xoredrows[1:]:
            rows.append(tuple(_r ^ _p for _r, _p in zip(row, rows[-1])))
        bitmap_strike = Raster.from_matrix(rows, inklevels=(False, True))
    # extract width from width/offset table
    # (do we need to consider the width table, if defined?)
    locs = tuple(_loc.offset for _loc in loc_table)
    glyphs = tuple(
        Glyph(bitmap_strike.crop(left=_offs, right=bitmap_strike.width-_next))
        for _offs, _next in zip(locs[:-1], locs[1:])
    )
    # add glyph metrics
    # scalable-width table
    if fontrec.fontType.has_width_table:
        glyphs = tuple(
            # fixed-point value, unsigned integer in the high-order byte
            # and a fractional part in the low-order byte
            _glyph.modify(scalable_width=f'{_we.width / 256:.2f}')
            for _glyph, _we in zip(glyphs, width_table)
        )
    # image-height table
    # > The Font Manager creates this table.
    # this appears to mean any stored contents may well be meaningless
    #
    # if fontrec.fontType.has_height_table:
    #     glyphs = tuple(
    #         _glyph.modify(image_height=_he.height, top_offset=_he.offset)
    #         for _glyph, _he in zip(glyphs, height_table)
    #     )
    # width & offset
    glyphs = tuple(
        _glyph.modify(wo_offset=_wo.offset, wo_width=_wo.width)
        for _glyph, _wo in zip(glyphs, wo_table)
    )
    return dict(
        glyphs=glyphs,
        fontrec=fontrec,
    )


# https://www.kreativekorp.com/swdownload/lisa/AppleLisaFontFormat.pdf
_COMPRESSED_HEADER = be.Struct(
    type='uint16',
    compressedLength='uint32',
    decompressedLength='uint32',
)

def _uncompress_nfnt(data, offset):
    """Decompress a compressed FONT/NFNT resource."""
    header = _COMPRESSED_HEADER.from_bytes(data, offset)
    offset += _COMPRESSED_HEADER.size
    payload = data[offset:offset+header.compressedLength]
    iter = reversed(payload)
    output = bytearray()
    for byte in iter:
        for bit in reversed(bytes_to_bits(bytes((byte,)))):
            if bit:
                output.append(0)
            else:
                try:
                    output.append(next(iter))
                except StopIteration:
                    break
    # bitmap rows still need to be XORed afterwards
    return bytes((data[0], data[1] ^ 0x80)) + bytes(reversed(output))


def convert_nfnt(properties, glyphs, fontrec):
    """Convert mac glyph metrics to monobit glyph metrics."""
    # the 'width' in the width/offset table is the pen advance
    # while the 'offset' is the (positive) offset after applying the
    # (positive or negative) 'kernMax' global offset
    #
    # since
    #   (glyph) advance_width == left_bearing + width + right_bearing
    # after this transformation we should have
    #   (glyph) advance_width == wo.width
    # which means
    #   (total) advance_width == wo.width - kernMax
    # since
    #   (total) advance_width == (font) left_bearing + glyph.advance_width + (font) right_bearing
    # and (font) left_bearing = -kernMax
    # we need to adjust for kernMax on both left and right bearings - it is an
    # offset only, not a tightening of the advance width
    if not glyphs:
        return Font()
    glyphs = tuple(
        _glyph.modify(
            left_bearing=_glyph.wo_offset + fontrec.kernMax,
            right_bearing=(
                _glyph.wo_width - _glyph.width
                - (_glyph.wo_offset + fontrec.kernMax)
            )
        )
        if _glyph.wo_width != 0xff and _glyph.wo_offset != 0xff else _glyph
        for _glyph in glyphs
    )
    # store glyph-name encoding table
    # do this before setting codepoint labels so we don't drop the tag on the 'missing' glyph
    encoding_table = properties.pop('encoding-table', None)
    if encoding_table:
        glyphs = tuple(
            _glyph.modify(tag=encoding_table.get(_glyph.codepoint, ''))
            for _glyph in glyphs
        )
    # codepoint labels
    labelled = [
        _glyph.modify(codepoint=(_codepoint,))
        for _codepoint, _glyph in enumerate(glyphs[:-1], start=fontrec.firstChar)
    ]
    # last glyph is the "missing" glyph
    labelled.append(glyphs[-1].modify(tag=MISSING_TAG))
    # drop undefined glyphs & their labels, so long as they're empty
    glyphs = tuple(
        _glyph for _glyph in labelled
        if (_glyph.wo_width != 0xff and _glyph.wo_offset != 0xff) or (_glyph.width and _glyph.height)
    )
    # drop mac glyph metrics
    # keep scalable_width
    glyphs = tuple(_glyph.drop('wo_offset', 'wo_width') for _glyph in glyphs)
    # store kerning table
    # last as this needs to refer to codepoint labels
    kerning_table = properties.pop('kerning-table', None)
    twos_complement = properties.pop('twos-complement', True)
    if kerning_table:
        # > Kerning distance. The kerning distance, in pixels, for the two glyphs
        # > at a point size of 1. This is a 16-bit fixed point value, with the
        # > integer part in the high-order 4 bits, and the fractional part in
        # > the low-order 12 bits. The Font Manager measures the distance in pixels
        # > and then multiplies it by the requested point size
        kern_table = sorted(
            (
                _entry.kernFirst, _entry.kernSecond,
                properties['point_size'] * fixed_to_float(
                    _entry.kernWidth, twos_complement=twos_complement
                )
            )
            for _entry in kerning_table
        )
        glyphs = tuple(
            _glyph.modify(right_kerning=KernTable({
                _right: f'{_width:.2f}'
                for _left, _right, _width in kern_table
                if _glyph.codepoint and _left == int(_glyph.codepoint)
            }))
            for _glyph in glyphs
        )
    # store properties
    properties.update({
        # not overridable; also seems incorrect for system fonts
        #'spacing': 'monospace' if fontrec.fontType.fixed_width else 'proportional',
        'default_char': MISSING_TAG,
        'ascent': fontrec.ascent,
        'descent': fontrec.descent,
        'line_height': fontrec.ascent + fontrec.descent + fontrec.leading,
        'shift_up': -fontrec.descent,
        # remove the kerning table and encoding table now stored in glyphs
        'kerning_table': None,
        'encoding_table': None,
        'source_format': properties.get('source_format', None) or 'NFNT',
    })
    return Font(glyphs, **properties)



###############################################################################
# NFNT writer

def create_nfnt(
        font, endian, ndescent_is_high, create_width_table, create_height_table,
        resample_encoding,
    ):
    """Create NFNT/FONT resource."""
    # subset to characters storable in NFNT
    font = subset_for_nfnt(font, resample_encoding)
    nfnt_data = convert_to_nfnt(
        font, endian, ndescent_is_high, create_width_table, create_height_table
    )
    data = nfnt_data_to_bytes(nfnt_data)
    return data, nfnt_data.owt_loc_high, nfnt_data.fbr_extent


def subset_for_nfnt(font, resample_encoding):
    """Subset to glyphs storable in NFNT and append default glyph."""
    default = font.get_default_glyph().modify(labels=(), tag=MISSING_TAG)
    is_mac_encoding = str(EncodingName(font.encoding)) in MAC_ENCODING_NAMES
    if (
            resample_encoding is NOT_SET
            and font.encoding and not is_mac_encoding
            and not font.encoding in ('raw', 'ascii')
        ):
        resample_encoding = 'mac-roman'
    if resample_encoding:
        # NFNT can only store encodings from a pre-defined list of 'scripts'
        # for fonts with other encodings, get glyphs corresponding to mac-roman
        font = font.resample(encoding=resample_encoding, missing=None)
    if font.encoding == 'ascii':
        font = font.subset(range(0, 128))
    else:
        # it's not clear to me if/how NFNT resources are used for MBCS
        # currently we just use the single-byte sector
        # this affects mac-japanese, mac-korean which have multibyte codepoints
        font = font.subset(range(0, 256))
    if not font.glyphs:
        raise UnsupportedError('No suitable characters for NFNT font')
    # add the default glyph at the end
    font = font.append(glyphs=(default,))
    return font


def _normalize_glyph(g, ink_bounds):
    """
    Trim the horizontal to glyph's ink bounds
    and expand the vertical to font's ink bounds
    in preparation for generating the font strike data
    """
    if not g:
        return None
    # shrink to fit
    g = g.reduce()
    return g.expand(
        bottom=g.shift_up - ink_bounds.bottom,
        top=ink_bounds.top-g.height-g.shift_up
    )


def _normalize_metrics(font):
    """Reduce to ink bounds horizontally, font ink bounds vertically."""
    glyphs = tuple(_normalize_glyph(_g, font.ink_bounds) for _g in font.glyphs)
    font = font.modify(glyphs)
    return font


def _calculate_nfnt_glyph_metrics(glyphs):
    """Calculate metrics for NFNT format."""
    # calculate maximum kerning (=most negative bearing), zero if all bearings positive
    # this equals the kernMax field of the fontRec
    kern_max = min(0, min(_g.left_bearing for _g in glyphs if _g))
    # add apple metrics to glyphs
    glyphs = tuple(
        (
            None if _g is None else
            _g.modify(wo_offset=_g.left_bearing-kern_max, wo_width=_g.advance_width)
        )
        for _g in glyphs
    )
    # check that glyph widths and offsets fit
    if any(_g.wo_width >= 255 for _g in glyphs if _g):
        raise UnsupportedError('NFNT character width must be < 255')
    if any(_g.wo_offset >= 255 for _g in glyphs if _g):
        raise UnsupportedError('NFNT character offset must be < 255')
    levels = max((_g.levels for _g in glyphs if _g), default=2)
    empty = Glyph(wo_offset=255, wo_width=255, levels=levels)
    glyphs = tuple(empty if _g is None else _g for _g in glyphs)
    return glyphs


def generate_nfnt_header(font, endian):
    """Generate a bare NFNT header with no bitmaps yet."""
    base = {'b': be, 'l': le}[endian[:1].lower()]
    NFNTHeader = nfnt_header_struct(base)
    FontType = NFNTHeader.element_types['fontType']
    # subset_for_nfnt has sorted on codepoint and added a 'missing' glyph
    first_char = int(min(font.get_codepoints()))
    last_char = int(max(font.get_codepoints()))
    # generate NFNT header
    depth = ((font.levels - 1).bit_length() -1).bit_length()
    fontrec = NFNTHeader(
        # this seems to be always 0x9000, 0xb000
        # even if docs say reserved_15 should be 0
        # we don't provide a scalable width or height tables
        fontType=FontType(
            reserved_15=1, reserved_12=1,
            depth=depth,
            fixed_width=font.spacing in ('monospace', 'character-cell'),
        ),
        firstChar=first_char,
        lastChar=last_char,
        widMax=font.max_width,
        # An integer value that specifies the distance from the font rectangle's glyph origin
        # to the left edge of the font rectangle, in pixels. If a glyph in the font kerns
        # to the left, the amount is represented as a negative number. If the glyph origin
        # lies on the left edge of the font rectangle, the value of the kernMax field is 0
        kernMax=min(0, min(_g.left_bearing for _g in font.glyphs)),
        nDescent=font.ink_bounds.bottom,
        # font rectangle == font bounding box
        fRectWidth=font.bounding_box.x,
        fRectHeight=font.bounding_box.y,
        # word offset to width/offset table
        # keep 0 for empty NFNT
        owTLoc=0,
        # docs define fRectHeight = ascent + descent
        # and generally suggest ascent and descent equal ink bounds
        # that's also monobit's *default* ascent & descent but is overridable
        ascent=font.ink_bounds.top,
        descent=-font.ink_bounds.bottom,
        # define leading in terms of bounding box, not pixel-height
        leading=font.line_height - font.bounding_box.y,
        # rowWords is 0 for empty NFNT, strike width in words for NFNT with bitmaps.
        rowWords=0,
    )
    logging.debug('NFNT header: %s', fontrec)
    return Props(
        fontrec=fontrec,
        font_strike=b'', loc_table=b'', wo_table=b'',
        width_table=b'', height_table=b'',
    )


def convert_to_nfnt(
        font, endian, ndescent_is_high, create_width_table, create_height_table
    ):
    """Convert monobit font to NFNT/FONT data structures."""
    # fontType is ignored
    # glyph-width table and image-height table not included
    base = {'b': be, 'l': le}[endian[:1].lower()]
    LocEntry = loc_entry_struct(base)
    WOEntry = wo_entry_struct(base)
    WidthEntry = width_entry_struct(base)
    HeightEntry = height_entry_struct(base)
    font = _normalize_metrics(font)
    # get contiguous glyph list
    # subset_for_nfnt has sorted on codepoint and added a 'missing' glyph
    if not font.get_codepoints():
        raise ValueError('No storable codepoints in font.')
    first_char = int(min(font.get_codepoints()))
    last_char = int(max(font.get_codepoints()))
    glyph_table = [
        font.get_glyph(codepoint=_code, missing=None)
        for _code in range(first_char, last_char+1)
    ]
    # reappend 'missing' glyph
    glyph_table.append(font.glyphs[-1])
    # calculate glyph metrics and fill in empties
    glyph_table = _calculate_nfnt_glyph_metrics(glyph_table)
    # build the font-strike data
    strike_raster = Raster.concatenate(*(_g.pixels for _g in glyph_table))
    # word-align strike
    strike_raster = strike_raster.expand(right=16-(strike_raster.width%16))
    font_strike = strike_raster.as_bytes(
        bits_per_pixel=(font.levels-1).bit_length()
    )
    # build the width-offset table
    empty = Glyph(wo_offset=255, wo_width=255, levels=font.levels)
    wo_table = b''.join(
        # glyph.wo_width and .wo_offset set in normalise_metrics
        bytes(WOEntry(width=_g.wo_width, offset=_g.wo_offset))
        # extra empty entry needed at the end.
        for _g in chain(glyph_table, [empty])
    )
    # build the location table
    loc_table = b''.join(
        bytes(LocEntry(offset=_offset))
        for _offset in accumulate((_g.width for _g in glyph_table), initial=0)
    )
    # build the glyph-width table
    if create_width_table:
        width_table = b''.join(
            bytes(WidthEntry(width=int(round(_g.scalable_width * 256))))
            for _g in glyph_table
        )
    else:
        width_table = b''
    # build the image-height table
    # this isn't tested and probably won't be - seems this table gets ignored
    if create_height_table:
        height_table = b''.join(
            bytes(HeightEntry(
                # offset from top line to first ink row
                # for normalised glyphs, this is the same as top padding
                offset=font.ink_bounds.top-_g.ink_bounds.top,
                height=_g.bounding_box.y
            ))
            for _g in glyph_table
        )
    else:
        height_table = b''
    # generate base fontrec
    fontrec = generate_nfnt_header(font, endian).fontrec
    fontrec.fontType.has_width_table = create_width_table
    fontrec.fontType.has_height_table = create_height_table
    # word offset to width/offset table
    # owTLoc is the offset from the field itself
    # the remaining size of the header including owTLoc is 5 words
    owt_loc = (len(font_strike) + len(loc_table) + 10) >> 1
    fontrec.owTLoc = owt_loc & 0xffff
    owt_loc_high = owt_loc >> 16
    if ndescent_is_high and owt_loc_high:
        fontrec.nDescent = owt_loc_high
    # fill in the rowWords, indicating that we do have a strike.
    pixels_per_byte = 8 // (font.levels - 1).bit_length()
    fontrec.rowWords = strike_raster.width // (2 * pixels_per_byte)
    # fbr = max width from origin (including whitespace) and right kerned pixels
    # for IIgs header
    fbr_extent = max(
        _g.width + _g.left_bearing + max(_g.right_bearing, 0)
        for _g in font.glyphs
    )
    return Props(
        fontrec=fontrec,
        font_strike=font_strike,
        loc_table=loc_table,
        wo_table=wo_table,
        width_table=width_table,
        height_table=height_table,
        owt_loc_high=owt_loc_high,
        fbr_extent=fbr_extent,
    )


def nfnt_data_to_bytes(nfnt_data):
    """Convert NFNT/FONT dtata structure to binary representation."""
    return b''.join((
        bytes(nfnt_data.fontrec),
        nfnt_data.font_strike,
        nfnt_data.loc_table,
        nfnt_data.wo_table,
        nfnt_data.width_table,
        nfnt_data.height_table,
    ))



"""
monobit - tools for working with monochrome bitmap fonts

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import sys as _sys
assert _sys.version_info >= (3, 9)

from .constants import VERSION as __version__
from .core import Pack, Font, Glyph, Char, Codepoint, Tag
from .base import FileFormatError, UnsupportedError
from .storage import load, save, loaders, savers
from .plumbing import scriptables as _operations
from .encoding import encoder, encodings
from .render import render, chart


# inject font operations into main module namespace
globals().update(_operations)

# make dash-versions of operations available through dict
operations = {
    _name.replace('_', '-'): _func
    for _name, _func in _operations.items()
}



"""
monobit.plumbing.args - script argument parser

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import os
import sys
import logging
from types import SimpleNamespace
from contextlib import contextmanager

ARG_PREFIX = '-'
GLOBAL_ARG_PREFIX = '--'
FALSE_PREFIX = 'no-'


class IsSetFlag:
    """
    Represent a parameter that is set with no value,
    converts to True or empty string, not to other types.
    """
    def __bool__(self):
        return True
    def __repr__(self):
        return f'{type(self).__name__}()'
    def __str__(self):
        return ''

SET = IsSetFlag()


def argrecord(command='', func=None, args=None, kwargs=None):
    """Record holding arguments and options for one command."""
    return SimpleNamespace(command=command, func=func, args=args or [], kwargs=kwargs or {})


def parse_subcommands(operations, global_options):
    """Split argument list in command components and their options with values."""
    # global arguments - these get added here wherever they occur in the argv list
    global_ns = argrecord()
    command_args = []
    for subargv in _split_argv(*operations.keys()):
        command = ''
        if subargv and subargv[0] in operations:
            command = subargv.pop(0)
        command_ns = argrecord(command=command, func=operations.get(command, ''))
        expect_value = False
        for arg in subargv:
            if arg.startswith((ARG_PREFIX, GLOBAL_ARG_PREFIX)):
                if arg.startswith(GLOBAL_ARG_PREFIX):
                    key = arg[len(GLOBAL_ARG_PREFIX):]
                    key, _, value = key.partition('=')
                    # use -- for global args
                    #  try to interpret as local if there isn't one
                    if key in global_options:
                        ns = global_ns
                    else:
                        ns = command_ns
                elif arg.startswith(ARG_PREFIX):
                    key = arg[len(ARG_PREFIX):]
                    key, _, value = key.partition('=')
                    ns = command_ns
                # boolean flag prefixed with --no-
                if key.startswith(FALSE_PREFIX):
                    key = key[len(FALSE_PREFIX):]
                    ns.kwargs[key] = False
                    expect_value = False
                else:
                    # record the key as set even if still expecting a value
                    ns.kwargs[key] = value or SET
                    expect_value = value == ''
            elif expect_value:
                value = arg
                ns.kwargs[key] = value
                expect_value = False
            else:
                # positional argument
                ns = command_ns
                ns.args.append(arg)
            # rename argument provided with dashes
            ns.kwargs = {
                _kwarg.replace('-', '_'): _value
                for _kwarg, _value in ns.kwargs.items()
            }
        command_args.append(command_ns)
    return command_args, global_ns


def _split_argv(*command_words):
    """Split argument list in command components."""
    part_argv = []
    for arg in sys.argv[1:]:
        if arg in command_words:
            yield part_argv
            part_argv = []
        part_argv.append(arg)
    yield part_argv


###############################################################################
# frame for main scripts

@contextmanager
def wrap_main(debug=False):
    """Main script context."""
    # set log level
    if debug:
        loglevel = logging.DEBUG
    else:
        loglevel = logging.WARNING
    logging.basicConfig(level=loglevel, format='%(levelname)s: %(message)s', force=True)
    # run main script
    try:
        yield
    except BrokenPipeError:
        # happens e.g. when piping to `head`
        # https://stackoverflow.com/questions/16314321/suppressing-printout-of-exception-ignored-message-in-python-3
        sys.stdout = os.fdopen(1)
    except Exception as exc:
        logging.error(exc)
        if debug:
            raise



"""
monobit.storage.fontformats.raw.comloaders - executable font loaders

(c) 2023--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging

from monobit.storage import loaders, Magic
from monobit.base import FileFormatError
from monobit.core import Glyph, Font, Char
from monobit.base.struct import little_endian as le, bitfield

from .raw import load_bitmap


###############################################################################
# REXXCOM Font Mania
# raw bitmap with DOS .COM header
# http://fileformats.archiveteam.org/wiki/Font_Mania_(REXXCOM)

# guessed by inspection, with reference to Intel 8086 opcodes
_FM_HEADER = le.Struct(
    # JMP SHORT opcode 0xEB
    jmp='uint8',
    # signed jump target - 0x4b or 0x4e
    code_offset='int8',
    bitmap_offset='uint16',
    bitmap_size='uint16',
    # seems to be always 0x2000 le, i.e. b'\0x20'.
    nul_space='2s',
    version_string='62s',
    # 'FONT MANIA, VERSION 1.0 \r\n COPYRIGHT (C) REXXCOM SYSTEMS, 1991'
    # 'FONT MANIA, VERSION 2.0 \r\n COPYRIGHT (C) REXXCOM SYSTEMS, 1991'
    # 'FONT MANIA, VERSION 2.2 \r\n COPYRIGHT (C) 1992  REXXCOM SYSTEMS'
)

@loaders.register(
    name='mania',
    magic=(Magic.offset(8) + b'FONT MANIA, VERSION',),
    patterns=('*.com',),
)
def load_mania(instream):
    """Load a REXXCOM Font Mania font."""
    header = _FM_HEADER.read_from(instream)
    logging.debug('Version string %r', header.version_string.decode('latin-1'))
    instream.read(header.bitmap_offset - _FM_HEADER.size)
    font = load_bitmap(
        instream,
        width=8, height=header.bitmap_size//256,
        count=256
    )
    font = font.modify(source_format='DOS loader (REXXCOM Font Mania)')
    return font


###############################################################################
# Fontraption
# raw bitmap with DOS .COM header, plain or TSR
# https://github.com/viler-int10h/Fontraption/blob/master/FORMATS.inc

_FRAPT_SIG = b'VILE\x1a'

_FRAPT_HEADER = le.Struct(
    magic='5s',
    loader_0='16s',
    height='uint8',
    loader_1='3s',
)
@loaders.register(
    name='frapt',
    magic=(_FRAPT_SIG,),
    patterns=('*.com',),
)
def load_frapt(instream):
    """Load a Fontraption plain .COM font."""
    header = _FRAPT_HEADER.read_from(instream)
    if header.magic != _FRAPT_SIG:
        raise FileFormatError(
            f'Not a Fontraption .COM file: incorrect signature {header.magic}.'
        )
    font = load_bitmap(instream, width=8, height=header.height)
    font = font.modify(source_format='DOS loader (Fontraption)')
    return font

@loaders.register(
    name='frapt-tsr',
    magic=(b'\xe9\x60',),
    patterns=('*.com',),
)
def load_frapt_tsr(instream):
    """Load a Fontraption TSR .COM font."""
    instream.seek(0x28)
    sig = instream.read(5)
    if sig != _FRAPT_SIG:
        raise FileFormatError(
            f'Not a Fontraption .COM file: incorrect signature {sig}.'
        )
    instream.seek(0x5d)
    height, = instream.read(1)
    instream.seek(0x63)
    font = load_bitmap(instream, width=8, height=height, count=256)
    font = font.modify(source_format='DOS TSR (Fontraption)')
    return font


###############################################################################
# FONTEDIT loader

_FONTEDIT_SIG = b'\xeb\x33\x90\r   \r\n PC Magazine \xfe Michael J. Mefford\0\x1a'

@loaders.register(
    name='fontedit',
    magic=(_FONTEDIT_SIG,),
    patterns=('*.com',),
)
def load_fontedit(instream):
    """Load a FONTEDIT .COM font."""
    sig = instream.read(99)
    if not sig.startswith(_FONTEDIT_SIG):
        raise FileFormatError(
            'Not a FONTEDIT .COM file: incorrect signature '
            f'{sig[:len(_FONTEDIT_SIG)]}.'
        )
    height = sig[50]
    font = load_bitmap(instream, width=8, height=height, count=256)
    font = font.modify(source_format='DOS loader (FONTEDIT)')
    return font


###############################################################################
# psftools PSF2AMS font loader
# raw bitmap with Z80 CP/M loader, prefixed with a DOS stub, 512-bytes offset
# https://github.com/ZXSpectrumVault/john-elliot/blob/master/psftools/tools/psf2ams.c
# /* Offsets in PSFCOM:
#  * 0000-000E  Initial code
#  * 000F-002D  Signature
#  * 002E-002F  Length of font, bytes (2k or 4k)
#  * 0030-0031  Address of font */

#_PSFCOM_STUB = bytes.fromhex('eb04 ebc3 ???? b409 ba32 01cd 21cd 20')
_PSFCOM_SIG08 = b'\rFont converted with PSF2AMS\r\n\032'
_PSFCOM_SIG16 = b'\rFont Converted with PSF2AMS\r\n\032'
_PSFCOM_HEADER = le.Struct(
    code='15s',
    sig='31s',
    bitmap_size='uint16',
    # apparently the offset to the space char, but 0-31 are defined before that
    # so this is the offset - 0x100 ?
    address='uint16',
)

@loaders.register(
    name='psfcom',
    magic=(
        b'\xeb\x04\xeb\xc3' + Magic.offset(11) + _PSFCOM_SIG08,
        b'\xeb\x04\xeb\xc3' + Magic.offset(11) + _PSFCOM_SIG16,
    ),
    patterns=('*.com',),
)
def load_psfcom(instream, first_codepoint:int=0):
    """
    Load a PSFCOM font.

    first_codepoint: first codepoint in file (default: 0)
    """
    header = _PSFCOM_HEADER.read_from(instream)
    logging.debug('Version string %r', header.sig.decode('latin-1'))
    if header.sig == _PSFCOM_SIG16:
        height = 16
    else:
        height = 8
    instream.read(header.address - _PSFCOM_HEADER.size - 0x100)
    font = load_bitmap(
        instream, width=8, height=height, first_codepoint=first_codepoint
    )
    font = font.modify(
        source_format='Amstrad/Spectrum CP/M loader (PSFCOM)',
        encoding='amstrad-cpm-plus',
    )
    font = font.label()
    return font


###############################################################################
# LETAFONT
# https://www.seasip.info/Unix/PSF/Amstrad/Letafont/index.html

@loaders.register(
    name='letafont',
    magic=(
        bytes.fromhex('210e 0111 90e2 0164 08ed b0c3 90e2 2a01'),
    ),
    patterns=('*.com',),
)
def load_letafont(instream):
    """Load a LETAFONT font."""
    instream.read(44)
    font = load_bitmap(
        instream, width=8, height=8, count=256
    )
    font = font.modify(
        source_format='Amstrad/Spectrum CP/M loader (LETAFONT)',
        encoding='amstrad-cpm-plus',
    )
    font = font.label()
    return font


###############################################################################
# UDG
# https://www.seasip.info/Unix/PSF/Amstrad/UDG/index.html

@loaders.register(
    name='udg',
    magic=(
        bytes.fromhex('210e 0111 2ce2 0134 08ed b0c3 2ce2 2a01'),
    ),
    patterns=('*.com',),
)
def load_udg(instream):
    """Load a UDG .COM font."""
    instream.read(114)
    font = load_bitmap(
        instream, width=8, height=8, count=256
    )
    font = font.modify(
        source_format='Amstrad/Spectrum CP/M loader (UDG)',
        encoding='amstrad-cpm-plus',
    )
    font = font.label()
    return font



"""
monobit.storage.fontformats.gfxfont - Adafruit GFX font headers

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from io import BytesIO
from itertools import accumulate

from monobit.storage import loaders, savers
from monobit.core import Glyph, Font
from monobit.base import Props, reverse_dict
from monobit.base.struct import little_endian as le
from monobit.base.binary import ceildiv
from monobit.encoding import EncodingName

from monobit.storage.utils.source import CCodeReader, CCodeWriter
from monobit.storage.utils.limitations import ensure_single, make_contiguous, ensure_charcell


_GLYPH_FIELDS = (
    'bitmapOffset', 'width', 'height', 'xAdvance', 'xOffset', 'yOffset',
)
_FONT_FIELDS = (
    'bitmap', 'glyph', 'first', 'last', 'yAdvance'
)

@loaders.register(
    name='gfxfont',
    # patterns=('*.h', '*.c'),
)
def load_gfxfont(
        instream,
        glyph_type:str='GFXglyph',
        font_type:str='GFXfont',
        bitmap_type:str='uint8_t'
    ):
    """
    Load fonts from Adafruit GFX font header.

    glyph_type: type identifier for glyph metrics (default: 'GFXglyph')
    font_type: type identifier for font metrics (default: 'GFXfont')
    bitmap_type: type identifier for bitmap data (default: 'uint8_t')
    """
    instream = instream.text
    identifier = ''
    data = {}
    glyph_metrics = {}
    font_metrics = {}
    for line in instream:
        line = CCodeReader.strip_line_comments(line, instream)
        if CCodeReader.assign in line:
            identifier, _, _ = line.partition(CCodeReader.assign)
            logging.debug('Found assignment to `%s`', identifier)
            if CCodeReader.delimiters[0] in line:
                identifier = CCodeReader.clean_identifier(identifier.strip().removesuffix(' PROGMEM'))
                coded_data = CCodeReader.read_array(instream, line)
                if font_type in line:
                    font_metrics[identifier] = CCodeReader.decode_struct(
                        coded_data, _FONT_FIELDS
                    )
                elif glyph_type in line:
                    glyph_metrics[identifier] = CCodeReader.decode_struct_array(
                        coded_data, _GLYPH_FIELDS
                    )
                elif bitmap_type in line:
                    data[identifier] = CCodeReader.decode_array(coded_data)
    fonts = []
    for name, metrics in font_metrics.items():
        glyphs = []
        # clip off cast
        _, _, bitmap_name = metrics.bitmap.partition(')')
        _, _, glyph_name = metrics.glyph.partition(')')
        first = CCodeReader.decode_int(metrics.first)
        try:
            glyph_table = glyph_metrics[glyph_name.strip()]
            bitmap = data[bitmap_name.strip()]
        except KeyError:
            logging.debug('Data for identifiers `%s` `%s` not found', glyph_name, bitmap_name)
            continue
        for cp, glyph_metric in enumerate(glyph_table, first):
            offset = CCodeReader.decode_int(glyph_metric.bitmapOffset)
            width = CCodeReader.decode_int(glyph_metric.width)
            height = CCodeReader.decode_int(glyph_metric.height)
            length = ceildiv(width * height, 8)
            x_advance = CCodeReader.decode_int(glyph_metric.xAdvance)
            x_offset = CCodeReader.decode_int(glyph_metric.xOffset)
            y_offset = CCodeReader.decode_int(glyph_metric.yOffset)
            glyphdata = bitmap[offset:offset+length]
            glyphs.append(Glyph.from_bytes(
                glyphdata, width=width, height=height, align='bit',
                left_bearing=x_offset,
                right_bearing=x_advance-x_offset-width,
                shift_up=-y_offset-height,
                codepoint=cp,
            ))
        fonts.append(Font(
            glyphs, line_height=CCodeReader.decode_int(metrics.yAdvance),
            name=CCodeReader.clean_identifier(name),
        ))
    return fonts


@savers.register(linked=load_gfxfont)
def save_gfxfont(fonts, outstream):
    """
    Save fonts to Adafruit GFX font header.
    """
    for font in fonts:
        font = make_contiguous(font, missing='space')
        codepoints = font.get_codepoints()
        if not codepoints:
            raise ValueError('No storable codepoints found in font.')
        font = font.subset(codepoints=codepoints)
        font = font.reduce()
        glyph_data = tuple(_g.as_bytes(align='bit') for _g in font.glyphs)
        offsets = accumulate((len(_d) for _d in glyph_data), initial=0)
        glyph_props = tuple(
            Props(
                bitmapOffset=str(_offset),
                width=str(_g.width),
                height=str(_g.height),
                xAdvance=str(_g.advance_width),
                xOffset=str(_g.left_bearing),
                yOffset=str(-_g.shift_up-_g.height),
            )
            for _g, _offset in zip(font.glyphs, offsets)
        )
        basename = CCodeWriter.to_identifier(font.name)
        bitmap_name = f'{basename}_bitmaps'
        glyph_name = f'{basename}_glyphs'
        font_props = Props(
            bitmap='(uint8_t *) ' + bitmap_name,
            glyph='(GFXGlyph *) ' + glyph_name,
            first=CCodeWriter.encode_int(int(min(codepoints))),
            last=CCodeWriter.encode_int(int(max(codepoints))),
            YAdvance=str(font.line_height),
        )
        outstream = outstream.text
        if font.get_comment():
            outstream.write(f'/*\n{font.get_comment()}\n*/\n\n')
        outstream.write(f'const uint8_t {bitmap_name}[] PROGMEM = ')
        outstream.write(
            CCodeWriter.encode_array(b''.join(glyph_data), bytes_per_line=8)
        )
        outstream.write(';\n\n')
        outstream.write(f'const GFXglyph {glyph_name}[] PROGMEM = ' '{\n')
        for i, props in enumerate(glyph_props):
            if i:
                outstream.write(',\n')
            outstream.write(CCodeWriter.indent)
            outstream.write(
                CCodeWriter.encode_struct(props, show_names=False, compact=True)
            )
        outstream.write('\n};\n\n')
        outstream.write(f'const GFXfont {basename} PROGMEM = ')
        outstream.write(CCodeWriter.encode_struct(font_props, show_names=False))
        outstream.write(';\n\n')



"""
monobit.storage.fontformats.raw - raw and nearly raw binary formats

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from monobit.base import import_all
import_all(__name__)

from .raw import load_bitmap, save_bitmap



"""
monobit.storage.fontformats.alto - Xerox Alto font formats

(c) 2023 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from itertools import count

from monobit.base.struct import bitfield, flag, big_endian as be
from monobit.storage import loaders, savers
from monobit.base import Props, FileFormatError, UnsupportedError
from monobit.core import Font, Glyph, Raster, Tag


@loaders.register(
    name='alto',
    patterns=('*.al',),
)
def load_alto(instream):
    """Load font from Xerox Alto screen font file."""
    header, al_glyphs =  _read_alto(instream)
    glyphs = _convert_glyphs(header, al_glyphs)
    return Font(glyphs, line_height=header.height)


@loaders.register(
    name='bitblt',
    patterns=('*.strike',),
)
def load_bitblt(instream):
    """Load font from Xerox Alto .strike file."""
    props, glyphs =  _read_bitblt(instream)
    return Font(glyphs, **props)


@loaders.register(
    name='prepress',
    patterns=('*.ac',),
)
def load_prepress(instream):
    """Load font from Xerox Alto PrePress .ac file."""
    props, glyphs =  _read_prepress(instream)
    return Font(glyphs, **props)


##############################################################################
# .AL "CONVERT" format
# http://www.bitsavers.org/pdf/xerox/alto/printing/AltoFontFormats_Oct1980.pdf

_AL_HEADER = be.Struct(
    height='uint16',
    proportional=bitfield('uint8', 1),
    baseline=bitfield('uint8', 7),
    maxWidth='uint8',
)

_XH_DATA = be.Struct(
    XW='uint16',
    HD='uint8',
    XH='uint8',
)

def _read_alto(instream):
    """Read Xerox Alto screen font file."""
    header = _AL_HEADER.read_from(instream)
    # character pointer table
    # pointers are self-referenced word offsets to the XW value in character structs
    pointers = []
    i = 0
    while True:
        here = instream.tell()
        # beak if we are at the first pointed-to location
        if pointers and here >= min(pointers):
            break
        pointer = be.uint16.read_from(instream)
        # break on null pointer
        if not int(pointer):
            break
        # record absolute position in file
        pointers.append(here + pointer*2)
    al_glyphs = tuple(
        _read_glyph(instream, header, pointer, cp)
        for cp, pointer in enumerate(pointers)
    )
    return header, al_glyphs


def _read_glyph(instream, header, pointer, cp):
    instream.seek(pointer)
    xh_data = _XH_DATA.read_from(instream)
    instream.seek(-_XH_DATA.size - xh_data.XH*2, 1)
    glyph_data = instream.read(xh_data.XH*2)
    return Props(
        pixels=Raster.from_bytes(glyph_data, width=16),
        **vars(xh_data)
    )


def _convert_glyphs(header, al_glyphs):
    glyphs = tuple(
        _convert_glyph(header, props)
        for props in al_glyphs
    )
    final_glyphs = []
    for cp, g in enumerate(glyphs[:256]):
        if g:
            while hasattr(g, 'extension'):
                ext = glyphs[g.extension]
                if g.shift_up > ext.shift_up:
                    g = g.expand(bottom=g.shift_up - ext.shift_up)
                else:
                    ext = ext.expand(bottom=ext.shift_up - g.shift_up)
                if g.height > ext.height:
                    ext = ext.expand(top=g.height - ext.height)
                else:
                    g = g.expand(top=ext.height - g.height)
                g = Glyph(
                    Raster.concatenate(g.pixels, ext.pixels),
                    right_bearing=ext.right_bearing, shift_up=ext.shift_up
                )
            final_glyphs.append(g.modify(codepoint=cp))
    return final_glyphs


def _convert_glyph(header, props):
    if (props.XW % 2):
        advance_width = (props.XW-1) // 2
        extension = None
    else:
        advance_width=16
        # index of extension char
        extension = props.XW // 2
    glyph_props = dict(
        right_bearing=-16 + advance_width,
        shift_up=header.baseline-props.HD-props.XH+1,
        extension=extension,
    )
    if not props.pixels:
        if props.HD or props.XW > 1:
            return Glyph.blank(width=16, height=props.HD, **glyph_props)
    else:
        return Glyph(props.pixels, width=16, **glyph_props)
    return None


##############################################################################
# "plainstrike" or "BITBLT" format

# a variant with a separate StrikeIndex file (.StrikeX) exists, but:
# > to the best of my knowledge, no one has ever used a StrikeIndex format

_FORMAT_STRUCT = be.Struct(
    # always =1, meaning 'new style'
    oneBit=bitfield('uint16', 1),
    # =1 means StrikeIndex, =0 otherwise
    index=bitfield('uint16', 1),
    # =1 if all characters have same value of Wx, else =0
    fixed=bitfield('uint16', 1),
    # =1 if KernedStrike, =0 if PlainStrike
    kerned=bitfield('uint16', 1),
    blank=bitfield('uint16', 12),
)

_STRIKE_HEADER = be.Struct(
    format=_FORMAT_STRUCT,
    # minimum character code
    min='uint16',
    # maximum character code
    max='uint16',
    # maximum spacing width of any character = max{Wx}
    maxwidth='uint16',
)

_STRIKE_BODY = be.Struct(
    # number of words in the StrikeBody
    length='uint16',
    # number of scan-lines above the baseline
    ascent='uint16',
    # number of scan-lines below the baseline
    descent='uint16',
    # always =0 [used to be used for padding schemes]
    xoffset='uint16',
    # number of words per scan-line in the strike
    raster='uint16',
    # the bit map, where height = ascent + descent
    # bitmap word raster*height
    # pointers into the strike, indexed by code
    # xinsegment ↑ min, max+2 word
)

_BOUNDING_BOX_BLOCK = be.Struct(
    # X offset
    FBBox='uint16',
    # Y offset
    FBBoy='uint16',
    # X width
    FBBdx='uint16',
    # Y width
    FBBdy='uint16',
)

_WIDTH_ENTRY = be.Struct(
    # the entire spacing word will be = (-1) (both bytes =377b)
    # to flag a non-existent character, else the bytes are:
    offset='int8',
    width='int8',
)

def _read_bitblt(instream):
    # @StrikeHeader
    header = _STRIKE_HEADER.read_from(instream)
    if not header.format.oneBit:
        raise FileFormatError('Not a Xerox BITBLT strike')
    if header.format.index:
        raise UnsupportedError('StrikeIndex format not supported')
    # @BoundingBoxBlock
    if header.format.kerned:
        bbox = _BOUNDING_BOX_BLOCK.read_from(instream)
    # @StrikeBody
    body = _STRIKE_BODY.read_from(instream)
    height = body.ascent + body.descent
    strikebytes = instream.read(2*body.raster*height)
    # max is the highest included code; max+1 holds the replacement char
    # (if min==max we have 2 glyphs and 3 bounds)
    offsets = (be.uint16 * (header.max+3 - header.min)).read_from(instream)
    # @WidthBody
    if header.format.kerned:
        spacing = (_WIDTH_ENTRY * (header.max+2 - header.min)).read_from(instream)
    strike = Raster.from_bytes(strikebytes, 16*body.raster, height)
    # conversion section
    # convert strike to glyphs
    if header.format.kerned:
        glyphs = [
            Glyph(
                strike.crop(left=_offset, right=max(0, 16*body.raster - _next)),
                codepoint=_cp,
                shift_up=-body.descent,
                left_bearing=_spacing.offset+bbox.FBBox,
                right_bearing=(
                    _spacing.width
                    -(_next-_offset)
                    -(_spacing.offset+bbox.FBBox)
                ),
            )
            for _offset, _next, _spacing, _cp in zip(
                offsets, offsets[1:], spacing, count(header.min)
            )
            if _spacing.width >= 0 and _spacing.offset >= 0
        ]
    else:
        glyphs = [
            Glyph(
                strike.crop(left=_offset, right=max(0, 16*body.raster - _next)),
                codepoint=_cp,
                shift_up=-body.descent,
            )
            for _offset, _next, _cp in zip(offsets, offsets[1:], count(header.min))
            if _offset != _next
        ]
    # last char is replacement char (referred to as "dummy character" in the docs)
    glyphs[-1] = glyphs[-1].modify(codepoint=None, tag='dummy')
    props = dict(
        ascent=body.ascent,
        descent=body.descent,
        default_char=Tag('dummy'),
    )
    return props, glyphs


##############################################################################
# PrePress format

_IX = be.Struct(
    # Various type codes are assigned
    type=bitfield('uint16', 4),
    # Length of entry in words, counting this one
    length=bitfield('uint16', 12),
)

# name-definition index entry
_IXN = be.Struct(
    # Header with type == 1
    ix=_IX,
    # The numeric code
    code='uint16',
    # The number of characters in the name
    nameLength='uint8',
    # Room for the name
    characters='19s',
)

_CHARACTER_INDEX_ENTRY = be.Struct(
    # type = 3 for character index
    ix=_IX,
    # Family name, using a name code
    family='uint8',
    # (if bold then 2 elseif light then 4 else 0) +
    # (if italic then 1 else 0) +
    # (if codensed then 6 elseif expanded then 12 else 0) +
    # (if Xerox then 0 elseif ASCII then 18 else 36)
    face='uint8',
    # Code for the "beginning character"
    bc='uint8',
    # Code for the "ending character"
    ec='uint8',
    # Size of the font segment [in micas == 10 micron units or 1/2540 in]
    size='uint16',
    # Rotation of the font segment
    rotation='uint16',
    # Starting address in file of the font segment
    segmentSA='uint32',
    # Length of the segment
    segmentLength='uint32',
    # Resolution in scan-lines/inch * 10
    resolutionX='uint16',
    # Resolution in bits/inch * 10
    resolutionY='uint16',
)


_CHARACTER_DATA = be.Struct(
    # X Width (scan-lines)
    # fixed-point, divide by 0x10000
    Wx='uint32',
    # Y width (bits)
    # fixed-point, divide by 0x10000
    Wy='uint32',
    # Bounding box offsets
    BBox='int16',
    BBoy='int16',
    # Width of bounding box (scan-lines)
    BBdx='uint16',
    # Height of bounding box (bits) or -1 for not defined
    BBdy='int16',
)

# relFilePos=uint32

_RASTER_DEFN = be.Struct(
    # Height of raster (in words)
    BBdyW=bitfield('uint16', 6),
    # Same as BBdx in CharacterData
    BBdx=bitfield('uint16', 10),
    # followed by raster, BBdyW*BBdx words
)


def _read_prepress(instream):
    """Read a PrePress file."""
    ixn = _IXN.read_from(instream)
    if ixn.ix.type != 1:
        raise FileFormatError('Not an .ac file: no name index entry')
    cie = _CHARACTER_INDEX_ENTRY.read_from(instream)
    if cie.ix.type != 3:
        raise FileFormatError('Not an .ac file: no character index entry')
    if cie.rotation != 0:
        raise UnsupportedError('Nonzero rotation not supported for this format.')
    instream.seek(cie.segmentSA*2)
    nchars = cie.ec - cie.bc + 1
    char_data = (_CHARACTER_DATA * nchars).read_from(instream)
    anchor = instream.tell()
    directory = (be.uint32 * nchars).read_from(instream)
    glyphs = []
    for cp, offset, cd in zip(range(cie.bc, cie.ec+1), directory, char_data):
        if cd.BBdy < 0:
            continue
        instream.seek(anchor+offset*2)
        raster_defn = _RASTER_DEFN.read_from(instream)
        raster = Raster.from_bytes(
            instream.read(2*raster_defn.BBdyW*raster_defn.BBdx),
            width=raster_defn.BBdyW*16,
            height=raster_defn.BBdx,
        ).turn(-1)
        glyphs.append(Glyph(
            raster, codepoint=cp,
            left_bearing=cd.BBox,
            shift_up=cd.BBoy,
            right_bearing=cd.Wx//0x10000-cd.BBox-cd.BBdx,
            # ignoring Wy - vertical advance
        ))
    # decode face number
    encoding = None
    setwidth = 'normal'
    face = cie.face
    if face >= 36:
        face -= 36
    elif face >= 18:
        encoding = 'ascii'
        face -= 18
    else:
        encoding = 'xerox'
    if face >= 12:
        setwidth = 'expanded'
        face -= 12
    elif face >= 6:
        setwidth = 'condensed'
        face -= 6
    weight = ('bold' if face&2 else 'light' if face&4 else 'regular')
    slant = ('italic' if face&1 else 'roman')
    # remaining properties
    props = dict(
        family=ixn.characters.decode('ascii', 'replace'),
        weight=weight,
        slant=slant,
        setwidth=setwidth,
        # encoding flag isn't set right in any of the .ac files seen
        #encoding=encoding,
        dpi=(cie.resolutionX//10, cie.resolutionY//10),
    )
    return props, glyphs
